import numpy as np
import numba

"""
real(kind=8) :: pi, tiny
real(kind=8), dimension(0:20) :: a, b, afar, a1, b1
real(kind=8), dimension(0:20) :: nrange
real(kind=8), dimension(0:20,0:20) :: gam
real(kind=8), dimension(8) :: xg, wg

initialize
----------
implicit none
real(kind=8) :: c, fac, twologhalf
real(kind=8), dimension(0:20) :: bot
real(kind=8), dimension(1:21) :: psi
integer :: n,m
"""

tiny = 1e-10
c = np.log(0.5) + 0.577215664901532860

fac = 1.0

nrange = np.arange(21, dtype=np.float_)

a = np.zeros(21, dtype=np.float_)
a[0] = 1.0
b = np.zeros(21, dtype=np.float_)

for n in range(1, 21):
    fac = n*fac
    a[n] = 1.0 / (4.0**nrange[n] * fac**2)
    b[n] = b[n-1] + 1 / nrange[n]

b = (b - c) * a
a = -a / 2.0

gam = np.zeros((21, 21), dtype=np.float_)
for n in range(21):
    for m in range(n+1):
        gam[n, m] = np.prod(nrange[m+1:n+1]) / np.prod(nrange[1:n-m+1])

afar = np.zeros(21, dtype=np.float_)
afar[0] = np.sqrt(np.pi / 2.)

for n in range(1, 21):
    afar[n] = -(2. * n - 1.)**2 / (n * 8) * afar[n-1]

fac = 1.0
bot = np.zeros(21, dtype=np.float_)
bot[0] = 4.0
for n in range(1, 21):
    fac = n * fac
    bot[n] = fac * (n+1)*fac * 4.0**(n+1)

psi = np.zeros(21, dtype=np.float_)
for n in range(2, 22):
    psi[n-1] = psi[n-2] + 1 / (n-1)
psi = psi - 0.577215664901532860

a1 = np.empty(21, dtype=np.float_)
b1 = np.empty(21, dtype=np.float_)
twologhalf = 2 * np.log(0.5)
for n in range(21):
    a1[n] = 1 / bot[n]
    b1[n] = (twologhalf - (2.0 * psi[n] + 1 / (n+1))) / bot[n]


wg = np.zeros(8, dtype=np.float_)
xg = np.zeros(8, dtype=np.float_)

wg[0] = 0.101228536290378
wg[1] = 0.22238103445338
wg[2] = 0.31370664587789
wg[3] = 0.36268378337836
wg[4] = 0.36268378337836
wg[5] = 0.313706645877890
wg[6] = 0.22238103445338
wg[7] = 0.10122853629038

xg[0] = -0.960289856497536
xg[1] = -0.796666477413626
xg[2] = -0.525532409916329
xg[3] = -0.183434642495650
xg[4] = 0.183434642495650
xg[5] = 0.525532409916329
xg[6] = 0.796666477413626
xg[7] = 0.960289856497536


@numba.njit(nogil=True, cache=True)
def besselk0far(z, Nt):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    integer, intent(in) :: Nt
    complex(kind=8) :: omega, term
    integer :: n

    """

    term = 1.0
    omega = afar[0]

    for n in range(1, Nt+1):
        term = term / z
        omega = omega + afar[n] * term

    omega = np.exp(-z) / np.sqrt(z) * omega

    return omega


@numba.njit(nogil=True, cache=True)
def besselk0near(z, Nt):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    integer, intent(in) :: Nt
    complex(kind=8) :: omega
    complex(kind=8) :: rsq, log1, term
    integer :: n
    """
    rsq = z**2
    term = 1.0 + 0.0j
    log1 = np.log(rsq)
    omega = a[0] * log1 + b[0]

    for n in range(1, Nt+1):
        term = term * rsq
        omega = omega + (a[n] * log1 + b[n]) * term

    return omega


@numba.njit(nogil=True, cache=True)
def besselk1near(z, Nt):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    integer, intent(in) :: Nt
    complex(kind=8) :: omega
    complex(kind=8) :: zsq, log1, term
    integer :: n
    """
    zsq = z**2
    term = z
    log1 = np.log(zsq)
    omega = 1. / z + (a1[0] * log1 + b1[0]) * z

    for n in range(1, Nt+1):
        term = term * zsq
        omega = omega + (a1[n] * log1 + b1[n]) * term

    return omega


@numba.njit(nogil=True, cache=True)
def besselk0cheb(z, Nt):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    integer, intent(in) :: Nt
    complex(kind=8) :: omega
    integer :: n, n2, ts
    real(kind=8) :: a, b, c, A3, u
    complex(kind=8) :: A1, A2, cn, cnp1, cnp2, cnp3
    complex(kind=8) :: z1, z2, S, T

    """
    cnp1 = complex(1., 0.)
    cnp2 = complex(0., 0.)
    cnp3 = complex(0., 0.)
    a = 0.5
    c = 1.
    b = 1. + a - c

    z1 = 2. * z
    z2 = 2. * z1
    ts = (-1)**(Nt+1)
    S = ts
    T = 1.

    for n in range(Nt, -1, -1):
        u = (n+a) * (n+b)
        n2 = 2 * n
        A1 = 1. - (z2 + (n2+3.)*(n+a+1.)*(n+b+1.) / (n2+4.)) / u
        A2 = 1. - (n2+2.)*(n2+3.-z2) / u
        A3 = -(n+1.)*(n+3.-a)*(n+3.-b) / (u*(n+2.))
        cn = (2.*n+2.) * A1 * cnp1 + A2 * cnp2 + A3 * cnp3
        ts = -ts
        S = S + ts * cn
        T = T + cn
        cnp3 = cnp2
        cnp2 = cnp1
        cnp1 = cn
    cn = cn / 2.
    S = S - cn
    T = T - cn
    omega = 1. / np.sqrt(z1) * T / S
    omega = np.sqrt(np.pi) * np.exp(-z) * omega

    return omega


@numba.njit(nogil=True, cache=True)
def besselk1cheb(z, Nt):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    integer, intent(in) :: Nt
    complex(kind=8) :: omega
    integer :: n, n2, ts
    real(kind=8) :: a, b, c, A3, u
    complex(kind=8) :: A1, A2, cn, cnp1, cnp2, cnp3
    complex(kind=8) :: z1, z2, S, T
    """
    cnp1 = 1.0 + 0.0j
    cnp2 = 0.0 + 0.0j
    cnp3 = 0.0 + 0.0j

    a = 1.5
    c = 3.0
    b = 1.0 + a - c

    z1 = 2 * z
    z2 = 2 * z1
    ts = (-1)**(Nt+1)
    S = ts
    T = 1.0

    for n in range(Nt, -1, -1):
        u = (n + a) * (n + b)
        n2 = 2 * n
        A1 = 1 - (z2 + (n2 + 3) * (n + a + 1) * (n + b + 1) / (n2 + 4)) / u
        A2 = 1 - (n2 + 2) * (n2 + 3 - z2) / u
        A3 = -(n + 1) * (n + 3 - a) * (n + 3 - b) / (u * (n + 2))
        cn = (2 * n + 2) * A1 * cnp1 + A2 * cnp2 + A3 * cnp3
        ts = -ts
        S = S + ts * cn
        T = T + cn
        cnp3 = cnp2
        cnp2 = cnp1
        cnp1 = cn

    cn = cn / 2
    S = S - cn
    T = T - cn
    omega = 1 / (np.sqrt(z1) * z1) * T / S
    omega = 2 * z * np.sqrt(np.pi) * np.exp(-z) * omega

    return omega


@numba.njit(nogil=True, cache=True)
def besselk0(x, y, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: z, omega
    real(kind=8) :: cond
    """
    z = np.sqrt(x**2 + y**2) / lab
    cond = np.abs(z)

    if (cond < 6):
        omega = besselk0near(z, 17)
    else:
        omega = besselk0cheb(z, 6)

    return omega


@numba.njit(nogil=True, cache=True)
def besselk1(x, y, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: z, omega
    real(kind=8) :: cond
    """
    z = np.sqrt(x**2 + y**2) / lab
    cond = np.abs(z)

    if (cond < 6):
        omega = besselk1near(z, 20)
    else:
        omega = besselk1cheb(z, 6)

    return omega


@numba.njit(nogil=True, cache=True)
def k0bessel(z):
    """
    implicit none
    complex(kind=8), intent(in) :: z
    complex(kind=8) :: omega
    real(kind=8) :: cond
    """
    cond = np.abs(z)

    if (cond < 6):
        omega = besselk0near(z, 17)
    else:
        omega = besselk0cheb(z, 6)

    return omega


@numba.njit(nogil=True, cache=True)
def besselk0v(x, y, lab, nlab, omega):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    integer, intent(in) :: nlab
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(nlab), intent(inout) :: omega
    integer :: n

    """
    for n in range(nlab):
        omega[n] = besselk0(x, y, lab[n])
    return omega


@numba.njit(nogil=True, cache=True)
def k0besselv(z, nlab, omega):
    """
    implicit none
    integer, intent(in) :: nlab
    complex(kind=8), dimension(nlab), intent(in) :: z
    complex(kind=8), dimension(nlab), intent(inout) :: omega
    integer :: n
    """
    for n in range(nlab):
        omega[n] = k0bessel(z[n])
    return omega


@numba.njit(nogil=True, cache=True)
def besselk0OLD(x, y, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: z, omega
    real(kind=8) :: cond
    """
    z = np.sqrt(x**2 + y**2) / lab
    cond = np.abs(z)

    if (cond < 4):
        omega = besselk0near(z, 12)  # Was 10
    elif (cond < 8):
        omega = besselk0near(z, 18)
    elif (cond < 12):
        omega = besselk0far(z, 11)  # was 6
    else:
        omega = besselk0far(z, 8)  # was 4

    return omega


# @numba.njit(nogil=True, cache=True)
# def besselcheb(z, Nt):
#     """
#     implicit none
#     integer, intent(in) :: Nt
#     complex(kind=8), intent(in) :: z
#     complex(kind=8) :: omega
#     complex(kind=8) :: z2
#     """
#     z2 = 2.0 * z
#     omega = np.sqrt(np.pi) * np.exp(-z) * ucheb(0.5, 1, z2, Nt)
#     return omega
# 
# 
# @numba.njit(nogil=True, cache=True)
# def ucheb(a, c, z, n0):
#     """
#     implicit none
#     integer, intent(in) :: c, n0
#     real(kind=8), intent(in) :: a
#     complex(kind=8), intent(in) :: z
#     complex(kind=8) :: ufunc
# 
#     integer :: n, n2, ts
#     real(kind=8) :: A3, u, b
#     complex(kind=8) :: A1, A2, cn,cnp1,cnp2,cnp3
#     complex(kind=8) :: z2, S, T
# 
#     """
# 
#     cnp1 = complex(1.)
#     cnp2 = complex(0.)
#     cnp3 = complex(0.)
#     ts = (-1)**(n0+1)
#     S = ts
#     T = 1.
#     z2 = 2. * z
#     b = 1. + a - c
# 
#     for n in range(n0, -1, -1):
#         u = (n+a) * (n+b)
#         n2 = 2 * n
#         A1 = 1. - (z2 + (n2+3)*(n+a+1)*(n+b+1.) / (n2+4.)) / u
#         A2 = 1. - (n2+2.)*(n2+3.-z2) / u
#         A3 = -(n+1)*(n+3-a)*(n+3-b) / (u*(n+2))
#         cn = (2*n+2) * A1 * cnp1 + A2 * cnp2 + A3 * cnp3
#         ts = -ts
#         S = S + ts * cn
#         T = T + cn
#         cnp3 = cnp2
#         cnp2 = cnp1
#         cnp1 = cn
#     cn = cn / 2.
#     S = S - cn
#     T = T - cn
#     ufunc = z**(-a) * T / S
#     return ufunc


# @numba.njit(nogil=True, cache=True)
# def besselk0complex(x, y):
#     """
#     implicit none
#     real(kind=8), intent(in) :: x,y
#     real(kind=8) :: phi
#     real(kind=8) :: d
#     complex(kind=8) :: zeta, zetabar, omega, logdminzdminzbar, dminzeta, term
#     complex(kind=8), dimension(0:20) :: zminzbar
#     complex(kind=8), dimension(0:20,0:20) :: gamnew
#     complex(kind=8), dimension(0:40) :: alpha, beta
#     """
#     d = 0.
#     zeta = complex(x, y)
#     zetabar = np.conj(zeta)
#     zminzbar = np.zeros(21, dtype=np.complex_)
#     for n in range(21):
#         # Ordered from high power to low power
#         zminzbar[n] = (zeta-zetabar)**(20-n)
# 
#     gamnew = np.asarray(gam, dtype=np.complex_)
#     for n in range(21):
#         gamnew[n, 0:n+1] = gamnew[n, 0:n+1] * zminzbar[20-n:20+1]
# 
#     alpha = np.zeros(41, dtype=np.complex_)
#     beta = np.zeros(41, dtype=np.complex_)
#     alpha[0] = a[0]
#     beta[0] = b[0]
#     for n in range(1, 21):
#         alpha[n:2*n+1] = alpha[n:2*n+1] + a[n] * gamnew[n, :n+1]
#         beta[n:2*n+1] = beta[n:2*n+1] + b[n] * gamnew[n, :n+1]
# 
#     omega = complex(0., 0.)
#     logdminzdminzbar = np.log((d-zeta) * (d-zetabar))
#     dminzeta = d - zeta
#     term = 1.
# 
#     for n in range(41):
#         omega = omega + (alpha[n] * logdminzdminzbar + beta[n]) * term
#         term = term * dminzeta
# 
#     phi = np.real(omega)
# 
#     return phi


@numba.njit(nogil=True, cache=True)
def lapls_int_ho(x, y, z1, z2, order):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(1:order+1) :: omega
    complex(kind=8), dimension(1:order+2) :: qm
    integer :: m, i
    real(kind=8) :: L
    complex(kind=8) :: z, zplus1, zmin1, log1, log2, log3, zpower
    """
    omega = np.empty((order+1,), dtype=np.complex_)
    qm = np.empty((order+2,), dtype=np.complex_)

    L = np.abs(z2-z1)
    z = (2 * complex(x, y) - (z1+z2)) / (z2-z1)
    zplus1 = z + 1
    zmin1 = z - 1

    # Not sure if this gives correct answer at corner point(z also appears in qm)
    # should really be caught in code that calls this def
    if (np.abs(zplus1) < tiny):
        zplus1 = tiny
    if (np.abs(zmin1) < tiny):
        zmin1 = tiny

    qm[0] = 0
    qm[1] = 2
    for m in range(2, order+1, 2):
        qm[m+1] = qm[m-1] * z * z + 2 / m

    for m in range(1, order+1, 2):
        qm[m+1] = qm[m] * z

    log1 = np.log((zmin1) / (zplus1))
    log2 = np.log(zmin1)
    log3 = np.log(zplus1)

    zpower = 1
    for i in range(1, order+2):
        zpower = zpower * z
        omega[i-1] = -L/(4*np.pi*i) * (zpower * log1 +
                                       qm[i] - log2 + (-1)**i * log3)

    return omega


@numba.njit(nogil=True, cache=True)
def bessellsreal(x, y, x1, y1, x2, y2, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y,x1,y1,x2,y2,lab
    real(kind=8) :: phi, biglab, biga, L
    complex(kind=8) :: z1, z2, zeta, zetabar, omega, log1, log2, term1, term2, d1minzeta, d2minzeta
    complex(kind=8), dimension(0:20) :: zminzbar
    complex(kind=8), dimension(0:20,0:20) :: gamnew
    complex(kind=8), dimension(0:40) :: alpha, beta

    """
    zminzbar = np.zeros(21, dtype=np.complex_)

    z1 = complex(x1, y1)
    z2 = complex(x2, y2)
    L = np.abs(z2-z1)
    biga = np.abs(lab)
    biglab = 2 * biga / L

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    for n in range(21):
        # Ordered from high power to low power
        zminzbar[n] = (zeta-zetabar)**(20-n)
    gamnew = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha[0] = a[0]
    beta[0] = b[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + a[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + b[n] * gamnew[n, 0:n+1]

    omega = 0
    d1minzeta = -1/biglab - zeta
    d2minzeta = 1/biglab - zeta
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)
    term1 = 1
    term2 = 1

    # I tried to serialize this, but it didn't speed things up
    for n in range(41):
        term1 = term1 * d1minzeta
        term2 = term2 * d2minzeta
        omega = omega + (2 * alpha[n] * log2 - 2 *
                         alpha[n] / (n+1) + beta[n]) * term2 / (n+1)
        omega = omega - (2 * alpha[n] * log1 - 2 *
                         alpha[n] / (n+1) + beta[n]) * term1 / (n+1)

    phi = -biga / (2*np.pi) * np.real(omega)

    return phi


@numba.njit(nogil=True, cache=True)
def bessellsrealho(x, y, x1, y1, x2, y2, lab, order):
    """
    implicit none
    real(kind=8), intent(in) :: x,y,x1,y1,x2,y2,lab
    integer, intent(in) :: order
    real(kind=8), dimension(0:order) :: phi
    real(kind=8) :: biglab, biga, L
    complex(kind=8) :: z1, z2, zeta, zetabar, omega, log1, log2, term1, term2, d1minzeta, d2minzeta
    complex(kind=8), dimension(0:20) :: zminzbar
    complex(kind=8), dimension(0:20,0:20) :: gamnew
    complex(kind=8), dimension(0:40) :: alpha, beta
    complex(kind=8) :: cm
    complex(kind=8), dimension(0:50) :: alphanew, betanew ! Maximum programmed order is 10
    integer :: n, m, p
    """
    phi = np.zeros(order+1, dtype=np.float_)
    zminzbar = np.zeros(21, dtype=np.complex_)

    z1 = complex(x1, y1)
    z2 = complex(x2, y2)
    L = np.abs(z2-z1)
    biga = np.abs(lab)
    biglab = 2 * biga / L

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    for n in range(21):
        # Ordered from high power to low power
        zminzbar[n] = (zeta-zetabar)**(20-n)
        
    gamnew = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha[0] = a[0]
    beta[0] = b[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + a[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + b[n] * gamnew[n, 0:n+1]

    d1minzeta = -1/biglab - zeta
    d2minzeta = 1/biglab - zeta
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(51, dtype=np.complex_)
    betanew = np.zeros(51, dtype=np.complex_)

    for p in range(order+1):
        alphanew[0:40+p] = 0
        betanew[0:40+p] = 0
        for m in range(p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]

        omega = 0
        term1 = 1
        term2 = 1
        # I tried to serialize this, but it didn't speed things up
        for n in range(40 + p + 1):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega = omega + \
                (2 * alphanew[n] * log2 - 2 * alphanew[n] /
                 (n+1) + betanew[n]) * term2 / (n+1)
            omega = omega - \
                (2 * alphanew[n] * log1 - 2 * alphanew[n] /
                 (n+1) + betanew[n]) * term1 / (n+1)

        phi[p] = -biga / (2*np.pi) * np.real(omega)

    return phi


@numba.njit(nogil=True, cache=True)
def bessells_int(x, y, z1, z2, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2,lab
    real(kind=8) :: biglab, biga, L, ang, tol
    complex(kind=8) :: zeta, zetabar, omega, log1, log2, term1, term2, d1minzeta, d2minzeta
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension(0:40) :: alpha, beta, alpha2
    integer :: n
    """
    zminzbar = np.zeros(21, dtype=np.complex_)

    L = np.abs(z2-z1)
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    biglab = 2 * biga / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a * exprange
    bnew = (b - a * complex(0, 2) * ang) * exprange

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    # #for n in range(21):
    # #    zminzbar[n] = (zeta-zetabar)**(20-n)  # Ordered from high power to low power
    zminzbar[20] = 1

    for n in range(1, 21):
        # Ordered from high power to low power
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)

    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)

    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]

    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    omega = 0
    d1minzeta = -1/biglab - zeta
    d2minzeta = 1/biglab - zeta

    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0)

    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)
    term1 = 1
    term2 = 1

    # I tried to serialize this, but it didn't speed things up
    for n in range(41):
        term1 = term1 * d1minzeta
        term2 = term2 * d2minzeta
        omega = omega + (alpha[n] * log2 - alpha[n] /
                         (n+1) + beta[n]) * term2 / (n+1)
        omega = omega - (alpha[n] * log1 - alpha[n] /
                         (n+1) + beta[n]) * term1 / (n+1)
        omega = omega + (alpha2[n] * np.conj(log2) -
                         alpha2[n] / (n+1)) * np.conj(term2) / (n+1)
        omega = omega - (alpha2[n] * np.conj(log1) -
                         alpha2[n] / (n+1)) * np.conj(term1) / (n+1)

    omega = -biga / (2*np.pi) * omega

    return omega


@numba.njit(nogil=True, cache=True)
def bessells_int_ho(x, y, z1, z2, lab, order, d1, d2):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:order) :: omega
    real(kind=8) :: biglab, biga, L, ang, tol
    complex(kind=8) :: zeta, zetabar, log1, log2, term1, term2, d1minzeta, d2minzeta, cm
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension(0:40) :: alpha, beta, alpha2
    complex(kind=8), dimension(0:50) :: alphanew, betanew, alphanew2 ! Order fixed to 10
    integer :: m, n, p
    """
    L = np.abs(z2-z1)
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    biglab = 2 * biga / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a * exprange
    bnew = (b - a * complex(0, 2) * ang) * exprange

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)

    # #for n in range(21):
    # #    zminzbar[n] = (zeta-zetabar)**(20-n)  # Ordered from high power to low power
    #
    zminzbar = np.zeros(21, dtype=np.complex_)
    zminzbar[20] = 1

    for n in range(1, 21):
        # Ordered from high power to low power
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)

    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)
    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    d1minzeta = d1/biglab - zeta
    d2minzeta = d2/biglab - zeta
    # #d1minzeta = -1/biglab - zeta
    # #d2minzeta = 1/biglab - zeta
    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0)
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(51, dtype=np.complex_)
    alphanew2 = np.zeros(51, dtype=np.complex_)
    betanew = np.zeros(51, dtype=np.complex_)

    omega = np.zeros(order+1, dtype=np.complex_)

    for p in range(order+1):
        alphanew[0:40+p+1] = 0
        betanew[0:40+p+1] = 0
        alphanew2[0:40+p+1] = 0

        for m in range(0, p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]
            cm = biglab**p * gam[p, m] * zetabar**(p-m)
            alphanew2[m:40+m+1] = alphanew2[m:40+m+1] + cm * alpha2[0:40+1]

        omega[p] = 0
        term1 = 1
        term2 = 1
        for n in range(41):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (alphanew[n] * log2 -
                                   alphanew[n] / (n+1) + betanew[n]) * term2 / (n+1)
            omega[p] = omega[p] - (alphanew[n] * log1 -
                                   alphanew[n] / (n+1) + betanew[n]) * term1 / (n+1)
            omega[p] = omega[p] + (alphanew2[n] * np.conj(log2) -
                                   alphanew2[n] / (n+1)) * np.conj(term2) / (n+1)
            omega[p] = omega[p] - (alphanew2[n] * np.conj(log1) -
                                   alphanew2[n] / (n+1)) * np.conj(term1) / (n+1)

    omega = -biga / (2*np.pi) * omega

    return omega


@numba.njit(nogil=True, cache=True)
def bessells_int_ho_qxqy(x, y, z1, z2, lab, order, d1, d2):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy
    complex(kind=8), dimension(0:order) :: qx, qy
    real(kind=8) :: biglab, biga, L, ang, angz, tol, bigx, bigy
    complex(kind=8) :: zeta, zetabar, log1, log2, term1, term2, d1minzeta, d2minzeta, bigz
    complex(kind=8) :: cm, biglabcomplex
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension(0:40) :: alpha, beta, alpha2
    complex(kind=8), dimension(0:51) :: alphanew, betanew, alphanew2 ! Order fixed to 10
    complex(kind=8), dimension(0:order+1) :: omega ! To store intermediate result
    complex(kind=8), dimension(0:order) :: omegalap ! To store intermediate result

    integer :: m, n, p
    """
    zminzbar = np.zeros(21, dtype=np.complex_)

    L = np.abs(z2-z1)
    bigz = (2 * complex(x, y) - (z1+z2)) / (z2-z1)
    bigx = bigz.real
    bigy = bigz.imag
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    angz = np.arctan2((z2-z1).imag, (z2-z1).real)
    biglab = 2 * biga / L
    biglabcomplex = 2.0 * lab / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a1 * exprange
    bnew = (b1 - a1 * complex(0, 2) * ang) * exprange

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    zminzbar[20] = 1

    for n in range(1, 21):
        # Ordered from high power to low po
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)
    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)

    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    d1minzeta = d1 / biglab - zeta
    d2minzeta = d2 / biglab - zeta

    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0)
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(52, dtype=np.complex_)
    alphanew2 = np.zeros(52, dtype=np.complex_)
    betanew = np.zeros(52, dtype=np.complex_)

    omega = np.zeros(order+2, dtype=np.complex_)
    qxqy = np.zeros(2*order+2, dtype=np.complex_)

    for p in range(0, order+2):

        alphanew[0:40+p+1] = 0
        betanew[0:40+p+1] = 0
        alphanew2[0:40+p+1] = 0
        for m in range(0, p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]
            cm = biglab**p * gam[p, m] * zetabar**(p-m)
            alphanew2[m:40+m+1] = alphanew2[m:40+m+1] + cm * alpha2[0:40+1]

        omega[p] = 0
        term1 = 1
        term2 = 1
        for n in range(40 + p + 1):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (alphanew[n] * log2 -
                                   alphanew[n] / (n+1) + betanew[n]) * term2 / (n+1)
            omega[p] = omega[p] - (alphanew[n] * log1 -
                                   alphanew[n] / (n+1) + betanew[n]) * term1 / (n+1)
            omega[p] = omega[p] + (alphanew2[n] * np.conj(log2) -
                                   alphanew2[n] / (n+1)) * np.conj(term2) / (n+1)
            omega[p] = omega[p] - (alphanew2[n] * np.conj(log1) -
                                   alphanew2[n] / (n+1)) * np.conj(term1) / (n+1)

    omega = biglab / (2*np.pi*biglabcomplex**2) * omega
    omegalap = lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2)

    # multiplication with 2/L inherently included
    qx = -(bigx * omega[0:order+1] - omega[1:order+2] + omegalap.imag)
    qy = -(bigy * omega[0:order+1] + omegalap.real)

    qxqy[0:order+1] = qx * np.cos(angz) - qy * np.sin(angz)
    qxqy[order+1:2*order+2] = qx * np.sin(angz) + qy * np.cos(angz)

    return qxqy


@numba.njit(nogil=True, cache=True)
def bessells_gauss(x, y, z1, z2, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: omega
    integer :: n
    real(kind=8) :: L, x0
    complex(kind=8) :: bigz, biglab
    """
    L = np.abs(z2-z1)
    biglab = 2 * lab / L
    bigz = (2 * complex(x, y) - (z1+z2)) / (z2-z1)
    omega = complex(0, 0)
    for n in range(1, 9):
        x0 = bigz.real - xg[n-1]
        omega = omega + wg[n-1] * besselk0(x0, bigz.imag, biglab)

    omega = -L/(4*np.pi) * omega
    return omega


@numba.njit(nogil=True, cache=True)
def bessells_gauss_ho(x, y, z1, z2, lab, order):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:order) :: omega
    integer :: n, p
    real(kind=8) :: L, x0
    complex(kind=8) :: bigz, biglab
    complex(kind=8), dimension(8) :: k0
    """
    L = np.abs(z2-z1)
    biglab = 2 * lab / L
    bigz = (2 * complex(x, y) - (z1+z2)) / (z2-z1)

    k0 = np.zeros(8, dtype=np.complex_)
    for n in range(8):
        x0 = bigz.real - xg[n]
        k0[n] = besselk0(x0, bigz.imag, biglab)

    omega = np.zeros(order+1, dtype=np.complex_)
    for p in range(order+1):
        omega[p] = complex(0, 0)
        for n in range(8):
            omega[p] = omega[p] + wg[n] * xg[n]**p * k0[n]
        omega[p] = -L/(4*np.pi) * omega[p]

    return omega


@numba.njit(nogil=True, cache=True)
def bessells_gauss_ho_d1d2(x, y, z1, z2, lab, order, d1, d2):
    """
    Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1

    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:order) :: omega, omegac
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """
    omega = np.zeros(order+1, dtype=np.complex_)

    bigz1 = complex(d1, 0)
    bigz2 = complex(d2, 0)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    omegac = bessells_gauss_ho(x, y, z1p, z2p, lab, order)
    dc = (d1+d2) / (d2-d1)
    for n in range(order+1):
        for m in range(n+1):
            omega[n] = omega[n] + gam[n, m] * dc**(n-m) * omegac[m]
        omega[n] = (0.5 * (d2-d1))**n * omega[n]
    return omega


@numba.njit(nogil=True, cache=True)
def bessells_gauss_ho_qxqy(x, y, z1, z2, lab, order):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy
    integer :: n, p
    real(kind=8) :: L, bigy, angz
    complex(kind=8) :: bigz, biglab
    real(kind=8), dimension(8) :: r, xmind
    complex(kind=8), dimension(8) :: k1
    complex(kind=8), dimension(0:order) :: qx,qy
    """
    qxqy = np.zeros(2*order+2, dtype=np.complex_)
    xmind = np.zeros(8, dtype=np.complex_)
    k1 = np.zeros(8, dtype=np.complex_)
    r = np.zeros(8, dtype=np.complex_)

    L = np.abs(z2-z1)
    biglab = 2 * lab / L
    bigz = (2 * complex(x, y) - (z1+z2)) / (z2-z1)
    bigy = bigz.imag
    for n in range(8):
        xmind[n] = bigz.real - xg[n]
        r[n] = np.sqrt(xmind[n]**2 + bigz.imag**2)
        k1[n] = besselk1(xmind[n], bigz.imag, biglab)

    qx = np.zeros(order+1, dtype=np.complex_)
    qy = np.zeros(order+1, dtype=np.complex_)
    for p in range(order+1):
        for n in range(8):
            qx[p] = qx[p] + wg[n] * xg[n]**p * xmind[n] * k1[n] / r[n]
            qy[p] = qy[p] + wg[n] * xg[n]**p * bigy * k1[n] / r[n]

    qx = -qx * L / (4*np.pi*biglab) * 2/L
    qy = -qy * L / (4*np.pi*biglab) * 2/L

    angz = np.arctan2((z2-z1).imag, (z2-z1).real)
    qxqy[0:order+1] = qx * np.cos(angz) - qy * np.sin(angz)
    qxqy[order+1:2*order+2] = qx * np.sin(angz) + qy * np.cos(angz)

    return qxqy


@numba.njit(nogil=True, cache=True)
def bessells_gauss_ho_qxqy_d1d2(x, y, z1, z2, lab, order, d1, d2):
    """
    Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1

    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy, qxqyc
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """
    qxqy = np.zeros(2*order+2, dtype=np.complex_)

    bigz1 = complex(d1, 0.)
    bigz2 = complex(d2, 0.)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    qxqyc = bessells_gauss_ho_qxqy(x, y, z1p, z2p, lab, order)
    dc = (d1+d2) / (d2-d1)
    for n in range(order+1):
        for m in range(n+1):
            qxqy[n] = qxqy[n] + gam[n, m] * dc**(n-m) * qxqyc[m]
            qxqy[n+order+1] = qxqy[n+order+1] + \
                gam[n, m] * dc**(n-m) * qxqyc[m+order+1]
        qxqy[n] = (0.5*(d2-d1))**n * qxqy[n]
        qxqy[n+order+1] = (0.5*(d2-d1))**n * qxqy[n+order+1]

    return qxqy


@numba.njit(nogil=True, cache=True)
def bessells(x, y, z1, z2, lab, order, d1in, d2in):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1in,d2in
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:order) :: omega

    integer :: Nls, n
    real(kind=8) :: Lnear, L, d1, d2, delta
    complex(kind=8) :: z, delz, za, zb
    """
    omega = np.zeros(order+1, dtype=np.complex_)

    Lnear = 3
    z = complex(x, y)
    L = np.abs(z2-z1)
    if (L < Lnear*np.abs(lab)):  # No need to break integral up
        if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
            omega = bessells_int_ho(x, y, z1, z2, lab, order, d1in, d2in)
        else:
            omega = bessells_gauss_ho_d1d2(
                x, y, z1, z2, lab, order, d1in, d2in)
    else:  # Break integral up in parts
        Nls = np.ceil(L / (Lnear*np.abs(lab)))
        delta = 2 / Nls
        delz = (z2-z1)/Nls
        L = np.abs(delz)
        for n in range(1, Nls+1):
            d1 = -1 + (n-1) * delta
            d2 = -1 + n * delta
            if ((d2 < d1in) or (d1 > d2in)):
                continue
            d1 = np.max(np.array([d1, d1in]))
            d2 = np.min(np.array([d2, d2in]))
            za = z1 + (n-1) * delz
            zb = z1 + n * delz
            if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                omega = omega + \
                    bessells_int_ho(x, y, z1, z2, lab, order, d1, d2)
            else:
                omega = omega + \
                    bessells_gauss_ho_d1d2(x, y, z1, z2, lab, order, d1, d2)
    return omega


@numba.njit(nogil=True, cache=True)
def bessellsv(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(nlab*(order+1)) :: omega
    integer :: n, nterms
    """
    nterms = order+1
    omega = np.zeros(nlab*(order+1), dtype=np.complex_)
    # Check if endpoints need to be adjusted using
    # the largest lambda (the first one)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(0, nlab):
        omega[n*nterms:(n+1)*nterms] = bessells(x, y, z1, z2, lab[n],
                                                  order, d1, d2)
    return omega


@numba.njit(nogil=True, cache=True)
def bessellsv2(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(order+1,nlab) :: omega
    integer :: n, nterms
    """
    nterms = order+1
    omega = np.zeros((order+1, nlab), dtype=np.complex_)
    # Check if endpoints need to be adjusted using the largest lambda (the first one)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        omega[:nterms+1, n] = bessells(x, y, z1, z2, lab[n], order, d1, d2)
    return omega


@numba.njit(nogil=True, cache=True)
def bessellsqxqy(x, y, z1, z2, lab, order, d1in, d2in):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1in,d2in
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy

    integer :: Nls, n
    real(kind=8) :: Lnear, L, d1, d2, delta
    complex(kind=8) :: z, delz, za, zb
    """
    Lnear = 3.
    z = complex(x, y)
    qxqy = np.zeros(2*order+2, dtype=np.complex_)
    L = np.abs(z2-z1)
    # print *,'Lnear*np.abs(lab) ',Lnear*np.abs(lab)
    if (L < Lnear*np.abs(lab)):  # No need to break integral up
        if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
            qxqy = bessells_int_ho_qxqy(x, y, z1, z2, lab, order, d1in, d2in)
        else:
            qxqy = bessells_gauss_ho_qxqy_d1d2(
                x, y, z1, z2, lab, order, d1in, d2in)

    else:  # Break integral up in parts
        Nls = np.ceil(L / (Lnear*np.abs(lab)))
        # print *,'NLS ',Nls
        delta = 2. / Nls
        delz = (z2-z1)/Nls
        L = np.abs(delz)
        for n in range(1, Nls+1):
            d1 = -1. + (n-1) * delta
            d2 = -1. + n * delta
            if ((d2 < d1in) or (d1 > d2in)):
                continue
            d1 = np.max(np.array([d1, d1in]))
            d2 = np.min(np.array([d2, d2in]))
            za = z1 + (n-1) * delz
            zb = z1 + n * delz
            if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                qxqy = qxqy + \
                    bessells_int_ho_qxqy(x, y, z1, z2, lab, order, d1, d2)
            else:
                qxqy = qxqy + \
                    bessells_gauss_ho_qxqy_d1d2(
                        x, y, z1, z2, lab, order, d1, d2)
    return qxqy


@numba.njit(nogil=True, cache=True)
def bessellsqxqyv(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(2*nlab*(order+1)) :: qxqy
    complex(kind=8), dimension(0:2*order+1) :: qxqylab
    integer :: n, nterms, nhalf
    """
    qxqy = np.zeros(2*nlab*(order+1), dtype=np.complex_)
    nterms = order+1
    nhalf = nlab*(order+1)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        qxqylab = bessellsqxqy(x, y, z1, z2, lab[n], order, d1, d2)
        qxqy[(n)*nterms:(n+1)*nterms] = qxqylab[0:order+1]
        qxqy[(n)*nterms+nhalf:(n+1)*nterms +
             nhalf] = qxqylab[order+1:2*order+1+1]
    return qxqy


@numba.njit(nogil=True, cache=True)
def bessellsqxqyv2(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(2*(order+1),nlab) :: qxqy
    complex(kind=8), dimension(0:2*order+1) :: qxqylab
    integer :: n, nterms, nhalf
    """
    qxqy = np.zeros((2*(order+1), nlab), dtype=np.complex_)
    nterms = order+1
    nhalf = nlab*(order+1)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        qxqylab = bessellsqxqy(x, y, z1, z2, lab[n-1], order, d1, d2)
        qxqy[:nterms, n] = qxqylab[0:order+1]
        qxqy[nterms:2*nterms, n] = qxqylab[order+1:2*order+1+1]
    return qxqy


@numba.njit(nogil=True, cache=True)
def bessellsuni(x, y, z1, z2, lab):
    """
    # Uniform strength
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: omega

    integer :: Nls, n
    real(kind=8) :: Lnear, L
    complex(kind=8) :: z, delz, za, zb
    """

    Lnear = 3.
    z = complex(x, y)
    omega = complex(0., 0.)
    L = np.abs(z2-z1)
    if (L < Lnear*np.abs(lab)):  # No need to break integral up
        if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
            omega = bessells_int(x, y, z1, z2, lab)
        else:
            omega = bessells_gauss(x, y, z1, z2, lab)
    else:  # Break integral up in parts
        Nls = np.ceil(L / (Lnear*np.abs(lab)))
        delz = (z2-z1)/Nls
        L = np.abs(delz)
        for n in range(1, Nls+1):
            za = z1 + (n-1) * delz
            zb = z1 + n * delz
            if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                omega = omega + bessells_int(x, y, za, zb, lab)
            else:
                omega = omega + bessells_gauss(x, y, za, zb, lab)
    return omega


@numba.njit(nogil=True, cache=True)
def bessellsuniv(x, y, z1, z2, lab, nlab):
    """
    # Uniform strength
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(nlab), intent(inout) :: omega
    integer :: n
    """
    omega = np.zeros(nlab, dtype=np.complex_)
    for n in range(nlab):
        omega[n] = bessellsuni(x, y, z1, z2, lab[n])
    return omega


# %% Line Doublet Functions
@numba.njit(nogil=True, cache=True)
def lapld_int_ho(x, y, z1, z2, order):
    """
    ! Near field only
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: omega, qm
    integer :: m, n
    real(kind=8) :: L
    complex(kind=8) :: z, zplus1, zmin1
    """

    omega = np.zeros(order+1, dtype=np.complex_)
    qm = np.zeros(order+1, dtype=np.complex_)

    L = np.abs(z2-z1)
    z = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    zplus1 = z + 1.
    zmin1 = z - 1.
    # Not sure if this gives correct answer at corner point (z also appears in qm); should really be caught in code that calls this function
    if np.abs(zplus1) < tiny:
        zplus1 = tiny
    if np.abs(zmin1) < tiny:
        zmin1 = tiny

    omega[0] = np.log(zmin1/zplus1)
    for n in range(1, order+1):
        omega[n] = z * omega[n-1]

    if order > 0:
        qm[1] = 2.
    for m in range(3, order+1, 2):
        qm[m] = qm[m-2] * z * z + 2. / m

    for m in range(2, order+1, 2):
        qm[m] = qm[m-1] * z

    omega = 1. / (complex(0., 2.) * np.pi) * (omega + qm)
    return omega


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2):
    """
    Near field only
    Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: omega, omegac
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """
    omega = np.zeros(order+1, dtype=np.complex_)

    bigz1 = complex(d1, 0.)
    bigz2 = complex(d2, 0.)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    omegac = lapld_int_ho(x, y, z1p, z2p, order)
    dc = (d1+d2) / (d2-d1)
    for n in range(order+1):
        for m in range(n+1):
            omega[n] = omega[n] + gam[n, m] * dc**(n-m) * omegac[m]
        omega[n] = (0.5*(d2-d1))**n * omega[n]

    return omega


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_wdis(x, y, z1, z2, order):
    """
    # Near field only
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: wdis
    complex(kind=8), dimension(0:10) :: qm  # Max order is 10
    integer :: m, n
    complex(kind=8) :: z, zplus1, zmin1, term1, term2, zterm
    """

    qm = np.zeros(11, dtype=np.complex_)
    wdis = np.zeros(order+1, dtype=np.complex_)

    z = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    zplus1 = z + 1.
    zmin1 = z - 1.
    # Not sure if this gives correct answer at corner point (z also appears in qm); should really be caught in code that calls this function
    if (np.abs(zplus1) < tiny):
        zplus1 = tiny
    if (np.abs(zmin1) < tiny):
        zmin1 = tiny

    qm[0:1] = 0.
    for m in range(2, order+1):
        qm[m] = 0.
        for n in range(1, m//2):
            qm[m] = qm[m] + (m-2*n+1) * z**(m-2*n) / (2*n-1)

    term1 = 1. / zmin1 - 1. / zplus1
    term2 = np.log(zmin1/zplus1)
    wdis[0] = term1
    zterm = complex(1., 0.)
    for m in range(1, order+1):
        wdis[m] = m * zterm * term2 + z * zterm * term1 + 2. * qm[m]
        zterm = zterm * z

    wdis = - wdis / (np.pi*complex(0., 1.)*(z2-z1))
    return wdis


@numba.njit(nogil=True, cache=True)
def lapld_int_ho_wdis_d1d2(x, y, z1, z2, order, d1, d2):
    """
    # Near field only
    # Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), dimension(0:order) :: wdis, wdisc
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """

    wdis = np.zeros(order+1, dtype=np.complex_)

    bigz1 = complex(d1, 0.)
    bigz2 = complex(d2, 0.)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    wdisc = lapld_int_ho_wdis(x, y, z1p, z2p, order)
    dc = (d1+d2) / (d2-d1)
    wdis[0:order+1] = 0.
    for n in range(order+1):
        for m in range(n+1):
            wdis[n] = wdis[n] + gam[n, m] * dc**(n-m) * wdisc[m]
        wdis[n] = (0.5*(d2-d1))**n * wdis[n]
    return wdis


@numba.njit(nogil=True, cache=True)
def besselld_int_ho(x, y, z1, z2, lab, order, d1, d2):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension[0:order+1] :: omega
    real(kind=8) :: biglab, biga, L, ang, tol, bigy
    complex(kind=8) :: zeta, zetabar, log1, log2, term1, term2, d1minzeta, d2minzeta, bigz
    complex(kind=8) :: cm, biglabcomplex
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension[0:40+1] :: alpha, beta, alpha2
    complex(kind=8), dimension(0:50) :: alphanew, betanew, alphanew2 # Order fixed to 10

    integer :: m, n, p
    """

    zminzbar = np.zeros(21, dtype=np.complex_)
    omega = np.zeros(order+1, dtype=np.complex_)

    L = np.abs(z2-z1)
    bigz = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    bigy = bigz.imag
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    biglab = 2. * biga / L
    biglabcomplex = 2.0 * lab / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a1 * exprange
    bnew = (b1 - a1 * complex(0, 2) * ang) * exprange

    zeta = (2. * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    zminzbar[20] = 1.
    for n in range(1, 21):
        # Ordered from high power to low power
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)

    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)

    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)
    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    d1minzeta = d1/biglab - zeta
    d2minzeta = d2/biglab - zeta
    #d1minzeta = -1./biglab - zeta
    #d2minzeta = 1./biglab - zeta
    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0.)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0.)
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(51, dtype=np.complex_)
    alphanew2 = np.zeros(51, dtype=np.complex_)
    betanew = np.zeros(51, dtype=np.complex_)

    for p in range(order+1):
        alphanew[0:40+p+1] = 0.
        betanew[0:40+p+1] = 0.
        alphanew2[0:40+p+1] = 0.
        for m in range(p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]
            cm = biglab**p * gam[p, m] * zetabar**(p-m)
            alphanew2[m:40+m+1] = alphanew2[m:40+m+1] + cm * alpha2[0:40+1]

        omega[p] = 0.
        term1 = 1.
        term2 = 1.
        for n in range(41):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (alphanew[n] * log2 -
                                   alphanew[n] / (n+1) + betanew[n]) * term2 / (n+1)
            omega[p] = omega[p] - (alphanew[n] * log1 -
                                   alphanew[n] / (n+1) + betanew[n]) * term1 / (n+1)
            omega[p] = omega[p] + (alphanew2[n] * np.conj(log2) -
                                   alphanew2[n] / (n+1)) * np.conj(term2) / (n+1)
            omega[p] = omega[p] - (alphanew2[n] * np.conj(log1) -
                                   alphanew2[n] / (n+1)) * np.conj(term1) / (n+1)

    # omega = bigy * biglab / (2.*pi*biglabcomplex**2) * omega + real( lapld_int_ho_d1d2(x,y,z1,z2,order,d1,d2) )
    omega = bigy * biglab / (2.*np.pi*biglabcomplex**2) * \
        omega + lapld_int_ho_d1d2(x, y, z1, z2, order, d1, d2).real
    return omega


@numba.njit(nogil=True, cache=True)
def besselld_gauss_ho(x, y, z1, z2, lab, order):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:order) :: omega
    integer :: n, p
    real(kind=8) :: L, x0, r
    complex(kind=8) :: bigz, biglab
    complex(kind=8), dimension(8) :: k1overr
    """

    k1overr = np.zeros(8, dtype=np.complex_)
    omega = np.zeros(order+1, dtype=np.complex_)

    L = np.abs(z2-z1)
    biglab = 2. * lab / L
    bigz = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    for n in range(8):
        x0 = bigz.real - xg[n]
        r = np.sqrt(x0**2 + bigz.imag**2)
        k1overr[n] = besselk1(x0, bigz.imag, biglab) / r
    for p in range(order+1):
        omega[p] = complex(0., 0.)
        for n in range(8):
            omega[p] = omega[p] + wg[n] * xg[n]**p * k1overr[n]

        omega[p] = bigz.imag/(2.*np.pi*biglab) * omega[p]
    return omega


@numba.njit(nogil=True, cache=True)
def besselld_gauss_ho_d1d2(x, y, z1, z2, lab, order, d1, d2):
    """
    # Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:order) :: omega, omegac
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """

    omega = np.zeros(order+1, dtype=np.complex_)

    bigz1 = complex(d1, 0.)
    bigz2 = complex(d2, 0.)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    omegac = besselld_gauss_ho(x, y, z1p, z2p, lab, order)
    dc = (d1+d2) / (d2-d1)
    omega[0:order+1] = 0.
    for n in range(order+1):
        for m in range(n+1):
            omega[n] = omega[n] + gam[n, m] * dc**(n-m) * omegac[m]
        omega[n] = (0.5*(d2-d1))**n * omega[n]
    return omega


@numba.njit(nogil=True, cache=True)
def besselld(x, y, z1, z2, lab, order, d1in, d2in):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1in,d2in
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:order) :: omega

    integer :: Nls, n
    real(kind=8) :: Lnear, L, d1, d2, delta
    complex(kind=8) :: z, delz, za, zb
    """

    omega = np.zeros(order+1, dtype=np.complex_)

    Lnear = 3.
    z = complex(x, y)
    L = np.abs(z2-z1)
    if (L < Lnear*np.abs(lab)):  # No need to break integral up
        if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
            omega = besselld_int_ho(x, y, z1, z2, lab, order, d1in, d2in)
        else:
            omega = besselld_gauss_ho_d1d2(
                x, y, z1, z2, lab, order, d1in, d2in)
    else:  # Break integral up in parts
        Nls = np.ceil(L / (Lnear*np.abs(lab)))
        delta = 2. / Nls
        delz = (z2-z1)/Nls
        L = np.abs(delz)
        for n in range(1, Nls+1):
            d1 = -1. + (n-1) * delta
            d2 = -1. + n * delta
            if ((d2 < d1in) or (d1 > d2in)):
                continue
            d1 = max(d1, d1in)
            d2 = min(d2, d2in)
            za = z1 + (n-1) * delz
            zb = z1 + n * delz
            if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                omega = omega + \
                    besselld_int_ho(x, y, z1, z2, lab, order, d1, d2)
            else:
                omega = omega + \
                    besselld_gauss_ho_d1d2(x, y, z1, z2, lab, order, d1, d2)

    return omega


@numba.njit(nogil=True, cache=True)
def besselldv(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(nlab*(order+1)) :: omega
    integer :: n, nterms
    """

    omega = np.zeros(nlab*(order+1), dtype=np.complex_)

    nterms = order+1
    # Check if endpoints need to be adjusted using the largest lambda (the first one)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        omega[(n)*nterms:(n+1)*nterms] = besselld(x, y, z1, z2,
                                                  lab[n], order, d1, d2)
    return omega


@numba.njit(nogil=True, cache=True)
def besselldv2(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(order+1,nlab) :: omega
    integer :: n, nterms
    """

    omega = np.zeros((order+1, nlab), dtype=np.complex_)

    nterms = order+1
    # Check if endpoints need to be adjusted using the largest lambda (the first one)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        omega[:nterms+1, n] = besselld(x, y, z1, z2, lab[n], order, d1, d2)

    return omega


# @numba.njit(nogil=True, cache=True)
def besselldpart(x, y, z1, z2, lab, order, d1, d2):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:order) :: omega
    real(kind=8) :: biglab, biga, L, ang, tol, bigy
    complex(kind=8) :: zeta, zetabar, log1, log2, term1, term2, d1minzeta, d2minzeta, bigz
    complex(kind=8) :: cm, biglabcomplex
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension(0:40) :: alpha, beta, alpha2
    complex(kind=8), dimension(0:50) :: alphanew, betanew, alphanew2 ! Order fixed to 10
    integer :: m, n, p
    """
    zminzbar = np.zeros(21, dtype=np.complex_)

    L = np.abs(z2-z1)
    bigz = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    bigy = bigz.imag
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    biglab = 2. * biga / L
    biglabcomplex = 2.0 * lab / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a1 * exprange
    bnew = (b1 - a1 * complex(0, 2) * ang) * exprange

    zeta = (2. * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    zminzbar[-1] = 1.
    for n in range(1, 21):
        # Ordered from high power to low power
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)

    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)
    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)
    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]
    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    d1minzeta = d1/biglab - zeta
    d2minzeta = d2/biglab - zeta

    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0.)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0.)
    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(51, dtype=np.complex_)
    alphanew2 = np.zeros(51, dtype=np.complex_)
    betanew = np.zeros(51, dtype=np.complex_)
    omega = np.zeros(order+1, dtype=np.complex_)

    for p in range(order+1):

        alphanew[0:40+p+1] = 0.
        betanew[0:40+p+1] = 0.
        alphanew2[0:40+p+1] = 0.
        for m in range(p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]
            cm = biglab**p * gam[p, m] * zetabar**(p-m)
            alphanew2[m:40+m+1] = alphanew2[m:40+m+1] + cm * alpha2[0:40+1]

        omega[p] = 0.
        term1 = 1.
        term2 = 1.
        for n in range(40+p+1):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (alphanew[n] * log2 -
                                   alphanew[n] / (n+1) + betanew[n]) * term2 / (n+1)
            omega[p] = omega[p] - (alphanew[n] * log1 -
                                   alphanew[n] / (n+1) + betanew[n]) * term1 / (n+1)
            omega[p] = omega[p] + (alphanew2[n] * np.conj(log2) -
                                   alphanew2[n] / (n+1)) * np.conj(term2) / (n+1)
            omega[p] = omega[p] - (alphanew2[n] * np.conj(log1) -
                                   alphanew2[n] / (n+1)) * np.conj(term1) / (n+1)

    # + real( lapld_int_ho(x,y,z1,z2,order) )
    omega = biglab / (2.*np.pi*biglabcomplex**2) * omega
    # omega = real( lapld_int_ho(x,y,z1,z2,order) )

    return omega


# @numba.njit(nogil=True, cache=True)
def besselld_int_ho_qxqy(x, y, z1, z2, lab, order, d1, d2):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy
    complex(kind=8), dimension(0:order) :: rvz, rvzbar
    real(kind=8) :: biglab, biga, L, ang, angz, tol, bigy
    complex(kind=8) :: zeta, zetabar, log1, log2, term1, term2, d1minzeta, d2minzeta, bigz
    complex(kind=8) :: cm, biglabcomplex, azero
    complex(kind=8), dimension(0:20) :: zminzbar, anew, bnew, exprange
    complex(kind=8), dimension(0:20,0:20) :: gamnew, gam2
    complex(kind=8), dimension(0:40) :: alpha, beta, alpha2
    complex(kind=8), dimension(0:51) :: alphanew, betanew, alphanew2 ! Order fixed to 10
    complex(kind=8), dimension(0:order) :: omegalap, omegaom, wdis, qx, qy ! To store intermediate result
    complex(kind=8), dimension(0:order+1) :: omega ! To store intermediate result
    integer :: m, n, p
    """

    zminzbar = np.zeros(21, dtype=np.complex_)
    omega = np.zeros(order+2, dtype=np.complex_)
    qxqy = np.zeros(2*order+2, dtype=np.complex_)

    L = np.abs(z2-z1)
    bigz = (2 * complex(x, y) - (z1+z2)) / (z2-z1)
    bigy = bigz.imag
    biga = np.abs(lab)
    ang = np.arctan2(lab.imag, lab.real)
    angz = np.arctan2((z2-z1).imag, (z2-z1).real)
    biglab = 2 * biga / L
    biglabcomplex = 2.0 * lab / L

    tol = 1e-12

    exprange = np.exp(-complex(0, 2) * ang * nrange)
    anew = a1 * exprange
    bnew = (b1 - a1 * complex(0, 2) * ang) * exprange
    azero = anew[0]

    for n in range(20):
        bnew[n] = (n+1)*bnew[n+1] + anew[n+1]
        anew[n] = (n+1)*anew[n+1]

    anew[20] = 0  # This is a bit lazy
    bnew[20] = 0

    zeta = (2 * complex(x, y) - (z1+z2)) / (z2-z1) / biglab
    zetabar = np.conj(zeta)
    zminzbar[20] = 1
    for n in range(1, 21):
        # Ordered from high power to low power
        zminzbar[20-n] = zminzbar[21-n] * (zeta-zetabar)

    gamnew = np.zeros((21, 21), dtype=np.complex_)
    gam2 = np.zeros((21, 21), dtype=np.complex_)

    for n in range(21):
        gamnew[n, 0:n+1] = gam[n, 0:n+1] * zminzbar[20-n:20+1]
        gam2[n, 0:n+1] = np.conj(gamnew[n, 0:n+1])

    alpha = np.zeros(41, dtype=np.complex_)
    alpha2 = np.zeros(41, dtype=np.complex_)
    beta = np.zeros(41, dtype=np.complex_)
    alpha[0] = anew[0]
    beta[0] = bnew[0]
    alpha2[0] = anew[0]

    for n in range(1, 21):
        alpha[n:2*n+1] = alpha[n:2*n+1] + anew[n] * gamnew[n, 0:n+1]
        beta[n:2*n+1] = beta[n:2*n+1] + bnew[n] * gamnew[n, 0:n+1]
        alpha2[n:2*n+1] = alpha2[n:2*n+1] + anew[n] * gam2[n, 0:n+1]

    d1minzeta = d1/biglab - zeta
    d2minzeta = d2/biglab - zeta
    # d1minzeta = -1/biglab - zeta
    # d2minzeta = 1/biglab - zeta
    if (np.abs(d1minzeta) < tol):
        d1minzeta = d1minzeta + complex(tol, 0)
    if (np.abs(d2minzeta) < tol):
        d2minzeta = d2minzeta + complex(tol, 0)

    log1 = np.log(d1minzeta)
    log2 = np.log(d2minzeta)

    alphanew = np.zeros(52, dtype=np.complex_)
    alphanew2 = np.zeros(52, dtype=np.complex_)
    betanew = np.zeros(52, dtype=np.complex_)

    for p in range(order+2):

        alphanew[0:40+p+1] = 0
        betanew[0:40+p+1] = 0
        alphanew2[0:40+p+1] = 0
        for m in range(p+1):
            cm = biglab**p * gam[p, m] * zeta**(p-m)
            alphanew[m:40+m+1] = alphanew[m:40+m+1] + cm * alpha[0:40+1]
            betanew[m:40+m+1] = betanew[m:40+m+1] + cm * beta[0:40+1]
            cm = biglab**p * gam[p, m] * zetabar**(p-m)
            alphanew2[m:40+m+1] = alphanew2[m:40+m+1] + cm * alpha2[0:40+1]

        omega[p] = 0
        term1 = 1
        term2 = 1
        for n in range(41):
            term1 = term1 * d1minzeta
            term2 = term2 * d2minzeta
            omega[p] = omega[p] + (alphanew[n] * log2 -
                                   alphanew[n] / (n+1) + betanew[n]) * term2 / (n+1)
            omega[p] = omega[p] - (alphanew[n] * log1 -
                                   alphanew[n] / (n+1) + betanew[n]) * term1 / (n+1)
            omega[p] = omega[p] + (alphanew2[n] * np.conj(log2) -
                                   alphanew2[n] / (n+1)) * np.conj(term2) / (n+1)
            omega[p] = omega[p] - (alphanew2[n] * np.conj(log1) -
                                   alphanew2[n] / (n+1)) * np.conj(term1) / (n+1)

    omegalap = lapld_int_ho_d1d2(
        x, y, z1, z2, order, d1, d2) / complex(0, 1)
    omegaom = besselldpart(x, y, z1, z2, lab, order, d1, d2)
    wdis = lapld_int_ho_wdis_d1d2(x, y, z1, z2, order, d1, d2)

    rvz = -biglab * bigy / (2*np.pi*biglabcomplex**2) * (omega[1:order+1+1]/biglab - zetabar * omega[0:order+1]) + \
        biglab * omegaom / complex(0, 2)
    rvzbar = -biglab * bigy / (2*np.pi*biglabcomplex**2) * (omega[1:order+1+1]/biglab - zeta * omega[0:order+1]) - \
        biglab * omegaom / complex(0, 2)
    # qxqy[0:order+1] = -2.0 / L * ( rvz + rvzbar ) / biglab  # As we need to take derivative w.r.t. z not zeta
    # qxqy[order+1:2*order+1+1] = -2.0 / L * complex(0,1) * (rvz-rvzbar) / biglab
    #
    # qxqy[0:order+1] = qxqy[0:order+1] - 2.0 / L / biglabcomplex**2 * azero * ( omegalap + np.conj(omegalap) )
    # qxqy[order+1:2*order+1+1] = qxqy[order+1:2*order+1+1] -  \
    #                          2.0 / L / biglabcomplex**2 * azero * complex(0,1) * (omegalap - np.conj(omegalap))
    #
    # qxqy[0:order+1] = qxqy[0:order+1] + real(wdis)
    # qxqy[order+1:2*order+1+1] = qxqy[order+1:2*order+1+1] - aimag(wdis)

    # As we need to take derivative w.r.t. z not zeta
    qx = -2.0 / L * (rvz + rvzbar) / biglab
    qy = -2.0 / L * complex(0, 1) * (rvz-rvzbar) / biglab

    qx = qx - 2.0 / L * bigy / biglabcomplex**2 * \
        azero * (omegalap + np.conj(omegalap))
    qy = qy - 2.0 / L * bigy / biglabcomplex**2 * azero * \
        complex(0, 1) * (omegalap - np.conj(omegalap))

    # qx = qx + real(wdis * (z2-z1) / L)
    # qy = qy - aimag(wdis * (z2-z1) / L)

    # print *,'angz ',angz
    # wdis already includes the correct rotation
    qxqy[0:order+1] = qx * np.cos(angz) - qy * np.sin(angz) + wdis.real
    qxqy[order+1:2*order+1+1] = qx * \
        np.sin(angz) + qy * np.cos(angz) - wdis.imag

    return qxqy


# @numba.njit(nogil=True, cache=True)
def besselld_gauss_ho_qxqy(x, y, z1, z2, lab, order):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy
    integer :: n, p
    real(kind=8) :: L, bigy, angz
    complex(kind=8) :: bigz, biglab
    real(kind=8), dimension(8) :: r, xmind
    complex(kind=8), dimension(8) :: k0,k1
    complex(kind=8), dimension(0:order) :: qx,qy
    """

    xmind = np.zeros(8, dtype=np.float_)
    r = np.zeros(8, dtype=np.float_)
    k0 = np.zeros(8, dtype=np.complex_)
    k1 = np.zeros(8, dtype=np.complex_)
    qxqy = np.zeros(2*order+2, dtype=np.complex_)

    L = np.abs(z2-z1)
    biglab = 2. * lab / L
    bigz = (2. * complex(x, y) - (z1+z2)) / (z2-z1)
    bigy = bigz.imag
    for n in range(8):
        xmind[n] = bigz.real - xg[n]
        r[n] = np.sqrt(xmind[n]**2 + bigz.imag**2)
        k0[n] = besselk0(xmind[n], bigz.imag, biglab)
        k1[n] = besselk1(xmind[n], bigz.imag, biglab)

    qx = np.zeros(order+1, dtype=np.complex_)
    qy = np.zeros(order+1, dtype=np.complex_)
    for p in range(order+1):
        for n in range(8):
            qx[p] = qx[p] + wg[n] * xg[n]**p * \
                (-bigy) * xmind[n] / r[n]**3 * \
                (r[n]*k0[n]/biglab + 2.*k1[n])
            qy[p] = qy[p] + wg[n] * xg[n]**p * \
                (k1[n]/r[n] - bigy**2 / r[n]**3 *
                 (r[n]*k0[n]/biglab + 2.*k1[n]))

    qx = -qx / (2*np.pi*biglab) * 2/L
    qy = -qy / (2*np.pi*biglab) * 2/L

    angz = np.arctan2((z2-z1).imag, (z2-z1).real)
    qxqy[0:order+1] = qx * np.cos(angz) - qy * np.sin(angz)
    qxqy[order+1:2*order+1+1] = qx * np.sin(angz) + qy * np.cos(angz)

    return qxqy


# @numba.njit(nogil=True, cache=True)
def besselld_gauss_ho_qxqy_d1d2(x, y, z1, z2, lab, order, d1, d2):
    """
    Returns integral from d1 to d2 along real axis while strength is still Delta^order from -1 to +1

    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1,d2
    complex(kind=8), intent(in) :: z1,z2,lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy, qxqyc
    integer :: n, m
    real(kind=8) :: xp, yp, dc, fac
    complex(kind=8) :: z1p,z2p,bigz1,bigz2
    """

    qxqy = np.zeros(2*order+2, dtype=np.complex_)

    bigz1 = complex(d1, 0)
    bigz2 = complex(d2, 0)
    z1p = 0.5 * (z2-z1) * bigz1 + 0.5 * (z1+z2)
    z2p = 0.5 * (z2-z1) * bigz2 + 0.5 * (z1+z2)
    qxqyc = besselld_gauss_ho_qxqy(x, y, z1p, z2p, lab, order)
    dc = (d1+d2) / (d2-d1)
    for n in range(order+1):
        for m in range(n+1):
            qxqy[n] = qxqy[n] + gam[n, m] * dc**(n-m) * qxqyc[m]
            qxqy[n+order+1] = qxqy[n+order+1] + \
                gam[n, m] * dc**(n-m) * qxqyc[m+order+1]

        qxqy[n] = (0.5*(d2-d1))**n * qxqy[n]
        qxqy[n+order+1] = (0.5*(d2-d1))**n * qxqy[n+order+1]

    return qxqy


# @numba.njit(nogil=True, cache=True)
def besselldqxqy(x, y, z1, z2, lab, order, d1in, d2in):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,d1in,d2in
    complex(kind=8), intent(in) :: z1,z2
    complex(kind=8), intent(in) :: lab
    complex(kind=8), dimension(0:2*order+1) :: qxqy

    integer :: Nls, n
    real(kind=8) :: Lnear, L, d1, d2, delta
    complex(kind=8) :: z, delz, za, zb

    """
    Lnear = 3
    z = complex(x, y)
    qxqy = complex(0, 0)
    L = np.abs(z2-z1)

    # print *,'Lnear*np.abs(lab) ',Lnear*np.abs(lab)
    if (L < Lnear*np.abs(lab)):  # No need to break integral up
        if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
            qxqy = besselld_int_ho_qxqy(x, y, z1, z2, lab, order, d1in, d2in)
        else:
            qxqy = besselld_gauss_ho_qxqy_d1d2(
                x, y, z1, z2, lab, order, d1in, d2in)

    else:  # Break integral up in parts
        Nls = np.ceil(L / (Lnear*np.abs(lab)))
        # print *,'NLS ',Nls
        delta = 2 / Nls
        delz = (z2-z1)/Nls
        L = np.abs(delz)
        for n in range(1, Nls+1):
            d1 = -1 + (n-1) * delta
            d2 = -1 + n * delta
            if ((d2 < d1in) or (d1 > d2in)):
                continue
            d1 = np.max(np.array([d1, d1in]))
            d2 = np.min(np.array([d2, d2in]))
            za = z1 + (n-1) * delz
            zb = z1 + n * delz
            if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                qxqy = qxqy + \
                    besselld_int_ho_qxqy(x, y, z1, z2, lab, order, d1, d2)
            else:
                qxqy = qxqy + \
                    besselld_gauss_ho_qxqy_d1d2(
                        x, y, z1, z2, lab, order, d1, d2)
    return qxqy


# @numba.njit(nogil=True, cache=True)
def besselldqxqyv(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(2*nlab*(order+1)) :: qxqy
    complex(kind=8), dimension(0:2*order+1) :: qxqylab
    """

    qxqy = np.zeros(2*nlab*(order+1), dtype=np.complex_)

    nterms = order+1
    nhalf = nlab*(order+1)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        qxqylab = besselldqxqy(x, y, z1, z2, lab[n], order, d1, d2)
        qxqy[n*nterms:(n+1)*nterms] = qxqylab[0:order+1]
        qxqy[n*nterms+nhalf:(n+1)*nterms +
             nhalf] = qxqylab[order+1:2*order+1+1]

    return qxqy


# @numba.njit(nogil=True, cache=True)
def besselldqxqyv2(x, y, z1, z2, lab, order, R, nlab):
    """
    implicit none
    integer, intent(in) :: order
    real(kind=8), intent(in) :: x,y,R
    complex(kind=8), intent(in) :: z1,z2
    integer, intent(in) :: nlab
    real(kind=8) :: d1, d2
    complex(kind=8), dimension(nlab), intent(in) :: lab
    complex(kind=8), dimension(2*(order+1),nlab) :: qxqy
    complex(kind=8), dimension(0:2*order+1) :: qxqylab
    integer :: n, nterms, nhalf
    """
    qxqy = np.zeros((2*(order+1), nlab), dtype=np.complex_)

    nterms = order+1
    nhalf = nlab*(order+1)
    d1, d2 = find_d1d2(z1, z2, complex(x, y), R*np.abs(lab[0]))
    for n in range(nlab):
        qxqylab = besselldqxqy(x, y, z1, z2, lab[n], order, d1, d2)
        qxqy[:nterms, n] = qxqylab[0:order+1]
        qxqy[nterms:2*nterms, n] = qxqylab[order+1:2*order+1+1]
    return qxqy


@numba.njit(nogil=True, cache=True)
def bessells_circcheck(x, y, z1in, z2in, lab):
    """
    implicit none
    real(kind=8), intent(in) :: x,y
    complex(kind=8), intent(in) :: z1in,z2in
    complex(kind=8), intent(in) :: lab
    complex(kind=8) :: omega

    integer :: Npt, Nls, n
    real(kind=8) :: Lnear, Lzero, L, x1, y1, x2, y2
    complex(kind=8) :: z, z1, z2, delz, za, zb
    """
    Lnear = 3
    Lzero = 20
    z = complex(x, y)
    x1, y1, x2, y2, Npt = circle_line_intersection(
        z1in, z2in, z, Lzero*np.abs(lab))

    z1 = complex(x1, y1)
    z2 = complex(x2, y2)

    omega = complex(0, 0)

    if (Npt == 2):
        L = np.abs(z2-z1)
        if (L < Lnear*np.abs(lab)):  # No need to break integral up
            if (np.abs(z - 0.5*(z1+z2)) < 0.5 * Lnear * L):  # Do integration
                omega = bessells_int(x, y, z1, z2, lab)
            else:
                omega = bessells_gauss(x, y, z1, z2, lab)
        else:  # Break integral up in parts
            Nls = np.ceil(L / (Lnear*np.abs(lab)))
            delz = (z2-z1)/Nls
            L = np.abs(delz)
            for n in range(1, Nls+1):
                za = z1 + (n-1) * delz
                zb = z1 + n * delz
                if (np.abs(z - 0.5*(za+zb)) < 0.5 * Lnear * L):  # Do integration
                    omega = omega + bessells_int(x, y, za, zb, lab)
                else:
                    omega = omega + bessells_gauss(x, y, za, zb, lab)
    return omega


@numba.njit(nogil=True, cache=True)
def is_too_far(z1, z2, zc, R):
    """
    Checks whether zc is more than R away from oval
    surrounding line element
    """
    
    Lover2 = np.abs(z2 - z1) / 2
    bigz = (2 * zc - (z1 + z2)) / (z2 - z1)
    Radj = R / Lover2
    
    rv = False
    if np.abs(bigz.imag) < Radj:
        if np.abs(bigz.real) < 1:
            rv = True
        elif np.abs(bigz.real) < 1 + Radj:
            if np.abs(zc - z1) < R:
                rv = True
            elif np.abs(zc - z2) < R:
                rv = True
    return rv

@numba.njit(nogil=True, cache=True)
def circle_line_intersection(z1, z2, zc, R):
    """
    implicit none
    complex(kind=8), intent(in) :: z1, z2, zc
    real(kind=8), intent(in) :: R
    real(kind=8), intent(inout) :: xouta, youta, xoutb, youtb
    integer, intent(inout) :: N
    real(kind=8) :: Lover2, d, xa, xb
    complex(kind=8) :: bigz, za, zb
    """
    N = 0
    za = complex(0, 0)
    zb = complex(0, 0)

    Lover2 = np.abs(z2-z1) / 2
    bigz = (2*zc - (z1+z2)) * Lover2 / (z2-z1)

    if (abs(bigz.imag) < R):
        d = np.sqrt(R**2 - bigz.imag**2)
        xa = bigz.real - d
        xb = bigz.real + d
        if ((xa < Lover2) and (xb > -Lover2)):
            N = 2
            if (xa < -Lover2):
                za = z1
            else:
                za = (xa * (z2-z1) / Lover2 + (z1+z2)) / 2.
            if (xb > Lover2):
                zb = z2
            else:
                zb = (xb * (z2-z1) / Lover2 + (z1+z2)) / 2.

    xouta = za.real
    youta = za.imag
    xoutb = zb.real
    youtb = zb.imag

    return xouta, youta, xoutb, youtb, N


@numba.njit(nogil=True, cache=True)
def find_d1d2(z1, z2, zc, R):
    """
    implicit none
    complex(kind=8), intent(in) :: z1, z2, zc
    real(kind=8), intent(in) :: R
    real(kind=8), intent(inout) :: d1, d2
    real(kind=8) :: Lover2, d, xa, xb
    complex(kind=8) :: bigz
    """
    d1 = -1.
    d2 = 1.

    Lover2 = np.abs(z2-z1) / 2
    bigz = (2*zc - (z1+z2)) * Lover2 / (z2-z1)

    if (np.abs((bigz.imag)) < R):
        d = np.sqrt(R**2 - bigz.imag**2)
        xa = bigz.real - d
        xb = bigz.real + d
        if ((xa < Lover2) and (xb > -Lover2)):
            if (xa < -Lover2):
                d1 = -1.
            else:
                d1 = xa / Lover2
            if (xb > Lover2):
                d2 = 1.
            else:
                d2 = xb / Lover2
    return d1, d2


@numba.njit(nogil=True, cache=True)
def isinside(z1, z2, zc, R):
    """ Checks whether point zc is within oval with 'radius' R from line element
    implicit none
    complex(kind=8), intent(in) :: z1, z2, zc
    real(kind=8), intent(in) :: R
    integer :: irv
    real(kind=8) :: Lover2, d, xa, xb
    complex(kind=8) :: bigz
    """
    irv = 0
    Lover2 = np.abs(z2-z1) / 2
    bigz = (2*zc - (z1+z2)) * np.abs(z2-z1) / (2*(z2-z1))

    if (np.abs(bigz.imag) < R):
        d = np.sqrt(R**2 - bigz.imag**2)
        xa = bigz.real - d
        xb = bigz.real + d
        if ((xa < Lover2) and (xb > -Lover2)):
            irv = 1
    return irv



import numpy as np
import matplotlib.pyplot as plt
import inspect # Used for storing the input
from .element import Element
from .equation import LeakyWallEquation
from . import besselnumba

class LineDoubletHoBase(Element):
    '''Higher Order LineDoublet Base Class. 
    All Higher Order Line Doublet elements are derived from this class'''
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, tsandbc=
                 [(0.0, 0.0)], res='imp', order=0, layers=0, type='', 
                 name='LineDoubletHoBase', label=None, addtomodel=True):
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers,
                         tsandbc=tsandbc, type=type, name=name, label=label)
        self.order = order
        self.nparam = (self.order + 1) * len(self.layers)
        self.x1 = float(x1)
        self.y1 = float(y1)
        self.x2 = float(x2)
        self.y2 = float(y2)
        if res == 'imp':
            self.res = np.inf
        else:
            self.res = float(res)
        if addtomodel: self.model.addelement(self)

    def __repr__(self):
        return self.name + ' from ' + str((self.x1, self.y1)) +\
                           ' to ' + str((self.x2, self.y2))

    def initialize(self):
        self.ncp = self.order + 1
        self.z1 = self.x1 + 1j * self.y1
        self.z2 = self.x2 + 1j * self.y2
        self.L = np.abs(self.z1 - self.z2)
        self.thetanormOut = np.arctan2(self.y2 - self.y1, self.x2 - self.x1) \
                            - np.pi / 2
        self.cosout = np.cos(self.thetanormOut) * np.ones(self.ncp)
        self.sinout = np.sin(self.thetanormOut) * np.ones(self.ncp)
        #
        thetacp = np.arange(np.pi, 0, -np.pi / self.ncp) \
                  - 0.5 * np.pi / self.ncp
        Zcp = np.zeros(self.ncp, 'D')
        Zcp.real = np.cos(thetacp)
        # control point just on positive site (this is handy later on)
        Zcp.imag = 1e-6  
        zcp = Zcp * (self.z2 - self.z1) / 2 + 0.5 * (self.z1 + self.z2)
        self.xc = zcp.real
        self.yc = zcp.imag
        # control point just on negative side 
        # (this is needed for building the system of equations)
        Zcp.imag = -1e-6  
        zcp = Zcp * (self.z2 - self.z1) / 2 + 0.5 * (self.z1 + self.z2)
        self.xcneg = zcp.real
        self.ycneg = zcp.imag  # control points just on negative side     
        #
        self.aq = self.model.aq.find_aquifer_data(self.xc[0], self.yc[0])
        self.setbc()
        coef = self.aq.coef[self.layers, :]
        self.setflowcoef()
        # shape (self.nlayers,self.aq.naq,self.model.npvalval)
        self.term = self.flowcoef * coef  
        self.term2 = self.term.reshape(self.nlayers, self.aq.naq, 
                                       self.model.nint, self.model.npint)
        self.resfac = self.aq.Haq[self.layers] / self.res
        self.dischargeinf = self.flowcoef * coef
        self.dischargeinflayers = np.sum(self.dischargeinf * 
                                         self.aq.eigvec[self.layers, :, :], 1)

    def setflowcoef(self):
        '''Separate function so that this can be overloaded for other types'''
        self.flowcoef = 1.0 / self.model.p  # Step function

    def potinf(self,x,y,aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.nint, 
                       self.model.npint), 'D')
        if aq == self.aq:
            pot = np.zeros((self.order+1,self.model.npint),'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    if besselnumba.isinside(self.z1, self.z2, x+y*1j, 
                                       self.rzero*self.aq.lababs[i, j]):
                        pot[:,:] = besselnumba.besselldv2(x, y, 
                            self.z1, self.z2, self.aq.lab2[i, j, :], 
                            self.order, 
                            self.rzero * self.aq.lababs[i, j]) / self.L 
                        for k in range(self.nlayers):
                            rv[k::self.nlayers, i, j, :] = \
                                self.term2[k, i, j, :] * pot
        rv.shape = (self.nparam, aq.naq, self.model.npval)
        return rv

    def disvecinf(self,x,y,aq=None):
        '''Can be called with only one x,y value'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros((self.nparam, aq.naq, self.model.nint, 
                        self.model.npint), 'D')
        rvy = np.zeros((self.nparam, aq.naq, self.model.nint, 
                        self.model.npint), 'D')
        if aq == self.aq:
            qxqy = np.zeros((2*(self.order+1),self.model.npint),'D')
            for i in range(self.aq.naq):
                for j in range(self.model.nint):
                    if besselnumba.isinside(self.z1, self.z2, x+y*1j, 
                                       self.rzero*self.aq.lababs[i, j]):
                        qxqy[:,:] = besselnumba.besselldqxqyv2(x, y, 
                            self.z1, self.z2,self.aq.lab2[i, j, :], 
                            self.order, 
                            self.rzero * self.aq.lababs[i, j]) / self.L
                        for k in range(self.nlayers):
                            rvx[k::self.nlayers, i, j, :] = \
                                self.term2[k, i, j, :] * \
                                qxqy[:self.order + 1,:]
                            rvy[k::self.nlayers, i, j, :] = \
                                self.term2[k, i, j, :] * \
                                qxqy[self.order + 1:,:]
                            
        rvx.shape = (self.nparam, aq.naq, self.model.npval)
        rvy.shape = (self.nparam, aq.naq, self.model.npval)
        return rvx, rvy

    def plot(self):
        plt.plot([self.x1, self.x2], [self.y1, self.y2], 'k')
    
class LeakyLineDoublet(LineDoubletHoBase, LeakyWallEquation):
    """
    Create a segment of a leaky wall, which is
    simulated with a line-doublet. The specific discharge through
    the wall is equal to the head difference across the wall
    divided by the resistance of the wall. 
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    x1 : scalar
        x-coordinate of fist point of line-doublet
    y1 : scalar
        y-coordinate of fist point of line-doublet
    x2 : scalar
        x-coordinate of second point of line-doublet
    y2 : scalar
        y-coordinate of second point of line-doublet
    res : scalar or string
        if string: 'imp' for an impermeable wall (same as res = np.inf)
        if scalar: resistance of leaky wall
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers 
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.LeakyLineDoubletString`
    
    """
    
    def __init__(self, model, x1=-1, y1=0, x2=1, y2=0, res='imp', order=0,
                 layers=0, label=None, addtomodel=True):
        self.storeinput(inspect.currentframe())
        LineDoubletHoBase.__init__(self, model, x1=x1, y1=y1, x2=x2, y2=y2,
                                   tsandbc=[(0, 0)], res=res, order=order,
                                   layers=layers, type='z', 
                                   name='LeakyLineDoublet', 
                                   label=label, addtomodel=addtomodel)
        self.nunknowns = self.nparam

    def initialize(self):
        LineDoubletHoBase.initialize(self)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        
class LeakyLineDoubletString(Element, LeakyWallEquation):
    """
    Create a string of leaky wall segements consisting
    of line-doublets
    
    Parameters
    ----------
    
    model : Model object
        Model to which the element is added
    xy : array or list
        list or array of (x,y) pairs of coordinates of end-points of
        the segements in the string
    res : scalar or string
        if string: 'imp' for an impermeable wall (same as res = np.inf)
        if scalar: resistance of leaky wall
    order : int (default is 0)
        polynomial order of potential jump along line-doublet
        (head jump if transmissivity is equal on each side of wall)
    layers : scalar, list or array
        layer(s) in which element is placed
        if scalar: element is placed in this layer
        if list or array: element is placed in all these layers
    label: str or None
        label of element
    
    See Also
    --------
    
    :class:`.LeakyLineDoublet`
    
    """
    
    def __init__(self, model, xy=[(-1, 0), (1, 0)], res='imp', order=0,
                 layers=0, label=None):
        self.storeinput(inspect.currentframe())
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=layers, 
                         tsandbc=[(0, 0)], type='z', 
                         name='LeakyLineDoubletString', label=label)
        self.res = res
        self.order = order
        self.ldlist = []
        xy = np.atleast_2d(xy).astype('d')
        self.x,self.y = xy[:,0], xy[:,1]
        self.nld = len(self.x) - 1
        for i in range(self.nld):
            self.ldlist.append(LeakyLineDoublet(model, 
                                   x1=self.x[i], y1=self.y[i],
                                   x2=self.x[i + 1], y2=self.y[i + 1],
                                   res=self.res, order=self.order,
                                   layers=layers, label=label,
                                   addtomodel=False))
        self.model.addelement(self)

    def __repr__(self):
        return self.name + ' with nodes ' + str(zip(self.x,self.y))

    def initialize(self):
        for ld in self.ldlist:
            ld.initialize()
        # Same order for all elements in string
        self.ncp = self.nld * self.ldlist[0].ncp  
        self.nparam = self.nld * self.ldlist[0].nparam
        self.nunknowns = self.nparam
        self.xld,self.yld = np.empty((self.nld,2)), np.empty((self.nld,2))
        for i,ld in enumerate(self.ldlist):
            self.xld[i,:] = [ld.x1,ld.x2]
            self.yld[i,:] = [ld.y1,ld.y2]
        # Only used for layout when it is a continuous string
        self.xldlayout = np.hstack((self.xld[:,0],self.xld[-1,1])) 
        self.yldlayout = np.hstack((self.yld[:,0],self.yld[-1,1]))
        self.aq = self.model.aq.find_aquifer_data(self.ldlist[0].xc, 
                                                  self.ldlist[0].yc)
        self.parameters = np.zeros((self.model.ngvbc, self.nparam, 
                                    self.model.npval), 'D')
        self.setbc()
        # As parameters are only stored for the element not the list,
        # we need to combine the following
        self.resfac = self.ldlist[0].resfac  # same for all elements in the list
        self.xc, self.yc = np.zeros(self.ncp), np.zeros(self.ncp)
        self.xcneg, self.ycneg = np.zeros(self.ncp), np.zeros(self.ncp)
        self.cosout, self.sinout = np.zeros(self.ncp), np.zeros(self.ncp)
        for i,ld in enumerate(self.ldlist):
            self.xc[i * ld.ncp: (i + 1) * ld.ncp] = ld.xc
            self.yc[i * ld.ncp: (i + 1) * ld.ncp] = ld.yc
            self.xcneg[i * ld.ncp: (i + 1) * ld.ncp] = ld.xcneg
            self.ycneg[i * ld.ncp: (i + 1) * ld.ncp] = ld.ycneg
            self.cosout[i * ld.ncp: (i + 1) * ld.ncp] = ld.cosout
            self.sinout[i * ld.ncp: (i + 1) * ld.ncp] = ld.sinout

    def potinf(self, x, y, aq=None):
        '''Returns array (nunknowns,nperiods)'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        for i,ld in enumerate(self.ldlist):
            rv[i*ld.nparam:(i+1)*ld.nparam,:] = ld.potinf(x,y,aq)
        return rv

    def disvecinf(self, x, y, aq=None):
        '''Returns array (nunknowns,nperiods)'''
        if aq is None: aq = self.model.aq.find_aquifer_data(x, y)
        rvx = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        rvy = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        for i,ld in enumerate(self.ldlist):
            qx,qy = ld.disvecinf(x,y,aq)
            rvx[i*ld.nparam:(i+1)*ld.nparam,:] = qx
            rvy[i*ld.nparam:(i+1)*ld.nparam,:] = qy
        return rvx,rvy
    
    def plot(self):
        plt.plot(self.xldlayout, self.yldlayout, 'k')



import numpy as np

def param_maq(kaq=[1], z=[1, 0], c=[], Saq=[0.001], Sll=[0], 
              poraq=[0.3], porll=[0.3], topboundary='conf', phreatictop=False):
    # Computes the parameters for a TimModel from input for a maq model
    z = np.atleast_1d(z).astype('d')
    kaq = np.atleast_1d(kaq).astype('d')
    Saq = np.atleast_1d(Saq).astype('d')
    poraq = np.atleast_1d(poraq).astype('d')
    c = np.atleast_1d(c).astype('d')
    Sll = np.atleast_1d(Sll).astype('d')
    porll = np.atleast_1d(porll).astype('d')
    H = z[:-1] - z[1:]
    assert np.all(H >= 0), 'Error: Not all layer thicknesses are' + \
                           ' non-negative' + str(H) 
    if topboundary[:3] == 'con':
        naq = int(len(z) / 2)
        if len(kaq) == 1:
            kaq = kaq * np.ones(naq)
        if len(Saq) == 1:
            Saq = Saq * np.ones(naq)
        if len(poraq) == 1:
            poraq = poraq * np.ones(naq)
        if len(c) == 1:
            c = c * np.ones(naq - 1)
        if len(Sll) == 1:
            Sll = Sll * np.ones(naq - 1)
        if len(porll) == 1:
            porll = porll * np.ones(naq - 1)
        assert len(kaq) == naq, 'Error: Length of kaq needs to be ' + \
                                  str(naq)
        assert len(Saq) == naq, 'Error: Length of Saq needs to be ' + \
                                str(naq)
        assert len(poraq) == naq, 'Error: Length of poraq needs to be ' + \
                                str(naq)
        assert len(c) == naq - 1, 'Error: Length of c needs to be ' + \
                                  str(naq - 1)
        assert len(Sll) == naq - 1, 'Error: Length of Sll needs to be ' + \
                                  str(naq - 1)
        assert len(porll) == naq - 1, 'Error: Length of porll needs to be ' + \
                                  str(naq - 1)
        Haq = H[::2]
        assert np.all(Haq > 0), 'Error: Some thicknesses of aquifer layers ' + \
                                'are negative'
        Hll = H[1::2]
        Hll = np.maximum(Hll, 1e-20) # make sure none are negative
        assert np.all(Hll > 0), 'Error: Some thicknesses of leaky layers ' + \
                                'are negative'
        c = np.hstack((1e100, c))  
        Sll = np.hstack((1e-20, Sll)) 
        Hll = np.hstack((1e-20, Hll))
        porll = np.hstack((1e-20, porll))
        # layertype
        nlayers = len(z) - 1
        ltype = np.array(nlayers * ['a'])
        ltype[1::2] = 'l'
    else: # leaky layers on top
        naq = int(len(z - 1) / 2)
        if len(kaq) == 1:
            kaq = kaq * np.ones(naq)
        if len(Saq) == 1:
            Saq = Saq * np.ones(naq)
        if len(poraq) == 1:
            poraq = poraq * np.ones(naq)
        if len(c) == 1:
            c = c * np.ones(naq)
        if len(Sll) == 1:
            Sll = Sll * np.ones(naq)
        if len(porll) == 1:
            porll = porll * np.ones(naq)
        assert len(kaq) == naq, 'Error: Length of kaq needs to be ' + \
                                  str(naq)
        assert len(Saq) == naq, 'Error: Length of Saq needs to be ' + \
                                str(naq)
        assert len(poraq) == naq, 'Error: Length of poraq needs to be ' + \
                                str(naq)
        assert len(c) == naq, 'Error: Length of c needs to be ' + \
                                  str(naq)
        assert len(Sll) == naq, 'Error: Length of Sll needs to be ' + \
                                  str(naq)
        assert len(porll) == naq, 'Error: Length of porll needs to be ' + \
                                  str(naq)
        Haq = H[1::2]
        Hll = H[::2]
        # layertype
        nlayers = len(z) - 1
        ltype = np.array(nlayers * ['a'])
        ltype[0::2] = 'l'
    return kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype
    
def param_3d(kaq=[1], z=[1, 0], Saq=[0.001], kzoverkh=1, poraq=0.3, 
             phreatictop=False, topboundary='conf', topres=0, topthick=0, 
             topSll=0, toppor=0.3):
    # Computes the parameters for a TimModel from input for a 3D model
    kaq = np.atleast_1d(kaq).astype('d')
    z = np.atleast_1d(z).astype('d')
    naq = len(z) - 1
    if len(kaq) == 1: 
        kaq = kaq * np.ones(naq)
    Saq = np.atleast_1d(Saq).astype('d')
    if len(Saq) == 1: 
        Saq = Saq * np.ones(naq)
    kzoverkh = np.atleast_1d(kzoverkh).astype('d')
    if len(kzoverkh) == 1: 
        kzoverkh = kzoverkh * np.ones(naq)
    poraq = np.atleast_1d(poraq).astype('d')
    if len(poraq) == 1:
        poraq = poraq * np.ones(naq)
    Haq = z[:-1] - z[1:]
    c = 0.5 * Haq[:-1] / (kzoverkh[:-1] * kaq[:-1]) + \
        0.5 * Haq[1:] /  (kzoverkh[1:] * kaq[1:])
    # Saq = Saq * H
    #if phreatictop:
    #    Saq[0] = Saq[0] / H[0]
    c = np.hstack((1e100, c))
    Hll = 1e-20 * np.ones(len(c))
    Sll = 1e-20 * np.ones(len(c))
    porll = np.zeros(len(c))
    nlayers = len(z) - 1
    ltype = np.array(nlayers * ['a'])
    if (topboundary[:3] == 'sem') or (topboundary[:3] == 'lea'):
        c[0] = np.max([1e-20, topres])
        Hll[0] = np.max([1e-20, topthick])
        Sll[0] = np.max([1e-20, topSll])
        porll[0] = toppor
        ltype = np.hstack(('l', ltype))
        z = np.hstack((z[0] + topthick, z))
    return kaq, Haq, Hll, c, Saq, Sll, poraq, porll, ltype, z


import numpy as np
import pandas as pd
from scipy.optimize import least_squares
import re
import lmfit

class Calibrate:
    
    def __init__(self, model):
        """initialize Calibration class
        
        Parameters
        ----------
        model : ttim.Model
            model to calibrate
        
        """

        self.model = model
        self.parameters = pd.DataFrame(columns=[
            'optimal', 'std', 'perc_std', 'pmin', 'pmax', 'initial', 'parray'])
        self.seriesdict = {}
        self.seriesinwelldict = {}
        
    def set_parameter(self, name=None, initial=0, pmin=-np.inf, pmax=np.inf):
        """set parameter to be optimized
        
        Parameters
        ----------
        name : str
            parameter name, can include layer information. 
            name can be 'kaq', 'Saq' or 'c'. A number after the parameter 
            name denotes the layer number, i.e. 'kaq0' refers to the hydraulic 
            conductivity of layer 0. 
            name also supports layer ranges, entered by adding a '_' and a
            layer number, i.e. 'kaq0_3' denotes conductivity for layers 0 up to 
            and including 3.
        initial : float, optional
            initial value for the parameter (the default is 0)
        pmin : float, optional
            lower bound for parameter value (the default is -np.inf)
        pmax : float, optional
            upper bound for paramater value (the default is np.inf)
        
        """

        assert type(name) == str, "Error: name must be string"
        # find numbers in name str for support layer ranges
        layers_from_name = re.findall(r'\d+', name)
        p = None
        if "_" in name:
            fromlay, tolay = [int(i) for i in layers_from_name]
            if name[:3] == 'kaq':
                p = self.model.aq.kaq[fromlay:tolay+1]
            elif name[:3] == 'Saq':
                p = self.model.aq.Saq[fromlay:tolay+1]
            elif name[0] == 'c':
                p = self.model.aq.c[fromlay:tolay+1]
            elif name[:3] == 'Sll':
                p = self.model.aq.Sll[fromlay:tolay+1]
            elif name[0:8] == 'kzoverkh':
                p = self.model.aq.kzoverkh[fromlay:tolay+1]
        else:
            layer = int(layers_from_name[0])
            # Set, kaq, Saq, c
            if name[:3] == 'kaq':
                p = self.model.aq.kaq[layer:layer + 1]
            elif name[:3] == 'Saq':
                p = self.model.aq.Saq[layer:layer + 1]
            elif name[0] == 'c':
                p = self.model.aq.c[layer:layer + 1]
            elif name[:3] == 'Sll':
                p = self.model.aq.Sll[layer:layer + 1]
            elif name[0:8] == 'kzoverkh':
                p = self.model.aq.kzoverkh[layer:layer + 1]
        if p is None:  # no parameter set
            print('parameter name not recognized or no parameter ref supplied')
            return
        self.parameters.loc[name] = {'optimal':initial, 'std':None, 
                                     'perc_std':None, 'pmin':pmin, 'pmax':pmax, 
                                     'initial':initial, 'parray':p[:]}
        
    def set_parameter_by_reference(self, name=None, parameter=None, initial=0, 
                                   pmin=-np.inf, pmax=np.inf):
        """set parameter to be optimized
        
        Parameters
        ----------
        name : str
            parameter name
        parameter : np.array
            array reference containing the parameter to be optimized. must be 
            specified as reference, i.e. w.rc[0:]  
        initial : float, optional
            initial value for the parameter (the default is 0)
        pmin : float, optional
            lower bound for parameter value (the default is -np.inf)
        pmax : float, optional
            upper bound for paramater value (the default is np.inf)
        
        """
        assert type(name) == str, "Error: name must be string"
        if parameter is not None:
            assert isinstance(parameter, np.ndarray), \
                "Error: parameter needs to be numpy array"
            p = parameter
        self.parameters.loc[name] = {'optimal':initial, 'std':None, 
                                     'perc_std':None, 'pmin':pmin, 'pmax':pmax, 
                                     'initial':initial, 'parray':p[:]}
        
    def series(self, name, x, y, layer, t, h):
        """method to add observations to Calibration object
        
        Parameters
        ----------
        name : str
            name of series
        x : float
            x-coordinate
        y : float
            y-coordinate
        layer : int
            layer number, 0-indexed
        t : np.array
            array containing timestamps of timeseries
        h : np.array
            array containing timeseries values, i.e. head observations
        
        """

        s = Series(x, y, layer, t, h)
        self.seriesdict[name] = s
        
    def seriesinwell(self, name, element, t, h):
        """method to add observations to Calibration object
        
        Parameters
        ----------
        name : str
            name of series
        element: element object with headinside function
        t : np.array
            array containing timestamps of timeseries
        h : np.array
            array containing timeseries values, i.e. head observations
        
        """

        e = SeriesInWell(element, t, h)
        self.seriesinwelldict[name] = e
        
    def residuals(self, p, printdot=False):
        """method to calculate residuals given certain parameters
        
        Parameters
        ----------
        p : np.array
            array containing parameter values
        printdot : bool, optional
            print dot for each function call
        
        Returns
        -------
        np.array
            array containing all residuals
        
        """

        if printdot:
            print('.', end='')
        # set the values of the variables
        
        if printdot == 7:
            print(p)
        
        for i, k in enumerate(self.parameters.index):
            # [:] needed to do set value in array
            self.parameters.loc[k, 'parray'][:] = p[i]  
            
        self.model.solve(silent=True)
        
        rv = np.empty(0)
        for key in self.seriesdict:
            s = self.seriesdict[key]
            h = self.model.head(s.x, s.y, s.t, layers=s.layer)
            rv = np.append(rv, s.h - h)
        for key in self.seriesinwelldict:
            s = self.seriesinwelldict[key]
            h = s.element.headinside(s.t)[0]
            rv = np.append(rv, s.h - h)
        return rv
    
    def residuals_lmfit(self, lmfitparams, printdot=False):
        vals = lmfitparams.valuesdict()
        p = np.array([vals[k] for k in self.parameters.index])
        #p = np.array([vals[k] for k in vals])
        return self.residuals(p, printdot)
    
    def fit_least_squares(self, report=True, diff_step=1e-4, xtol=1e-8, 
                          method='lm'):
        self.fitresult = least_squares(
            self.residuals, self.parameters.initial.values, args=(True,), 
            bounds=(self.parameters.pmin.values, self.parameters.pmax.values),
            method=method, diff_step=diff_step, xtol=xtol, x_scale="jac")
        print('', flush=True)
        # Call residuals to specify optimal values for model
        res = self.residuals(self.fitresult.x)
        for ipar in self.parameters.index:
            self.parameters.loc[ipar, 'optimal'] = \
                self.parameters.loc[ipar, 'parray'][0]
        nparam = len(self.fitresult.x)
        H = self.fitresult.jac.T @ self.fitresult.jac
        sigsq = np.var(res, ddof=nparam)
        self.covmat = np.linalg.inv(H) * sigsq 
        self.sig = np.sqrt(np.diag(self.covmat))
        D = np.diag(1 / self.sig)
        self.cormat = D @ self.covmat @ D
        self.parameters['std'] = self.sig
        self.parameters['perc_std'] = self.sig / \
                                      self.parameters['optimal'] * 100
        if report:
            print(self.parameters)
            print(self.sig)
            print(self.covmat)
            print(self.cormat)
            
    def fit_lmfit(self, report=True, printdot=True):
        import lmfit
        self.lmfitparams = lmfit.Parameters()
        for name in self.parameters.index:
            p = self.parameters.loc[name]
            self.lmfitparams.add(name, value=p['initial'], min=p['pmin'], 
                                 max=p['pmax'])
        fit_kws = {"epsfcn": 1e-4}
        self.fitresult = lmfit.minimize(self.residuals_lmfit, self.lmfitparams, 
                                        method="leastsq", 
                                        kws={"printdot":printdot}, **fit_kws)
        print('', flush=True)
        print(self.fitresult.message)
        if self.fitresult.success:
            for name in self.parameters.index:
                self.parameters.loc[name, 'optimal'] = \
                    self.fitresult.params.valuesdict()[name]
            if hasattr(self.fitresult, 'covar'):
                self.parameters['std'] = np.sqrt(np.diag(self.fitresult.covar))
                self.parameters['perc_std'] = 100 * self.parameters['std'] / \
                                              np.abs(self.parameters['optimal'])
            else:
                self.parameters['std'] = np.nan
                self.parameters['perc_std'] = np.nan
        if report:
            print(lmfit.fit_report(self.fitresult))
            
    def fit(self, report=True, printdot=True):
        # current default fitting routine
        return self.fit_lmfit(report, printdot)
            
    def rmse(self):
        """calculate root-mean-squared-error
        
        Returns
        -------
        float
            return rmse value
        """

        r = self.residuals(self.parameters['optimal'].values)
        return np.sqrt(np.mean(r ** 2))
    
class Series:
    def __init__(self, x, y, layer, t, h):
        self.x = x
        self.y = y
        self.layer = layer
        self.t = t
        self.h = h
        
class SeriesInWell:
    def __init__(self, element, t, h):
        self.element = element
        self.t = t
        self.h = h





import numpy as np
import matplotlib.pyplot as plt
from scipy.special import kv, iv
import inspect # Used for storing the input
from .element import Element

class CircAreaSink(Element):
    """
    Create a circular area-sink with uniform infiltration rate in aquifer
    layer 0.
    Infiltration rate in length / time, positive for water entering
    the aquifer.
    
    Parameters
    ----------
    model : Model object
        model to which the element is added
    xc : float
        x-coordinate of center of area-sink
    yc : float
        y-coordinate of center of area-sink
    R : radius of area-sink
    tsandN : list of tuples
        tuples of starting time and infiltration rate after starting time
    label : string or None (default: None)
        label of the area-sink
            
    """
    
    def __init__(self, model, xc=0, yc=0, R=0.1, tsandN=[(0, 1)],
                 name='CircAreaSink', label=None):
        self.storeinput(inspect.currentframe())
        Element.__init__(self, model, nparam=1, nunknowns=0, layers=0,
                         tsandbc=tsandN, type='g', name=name, label=label)
        self.xc = float(xc); self.yc = float(yc); self.R = float(R)
        self.model.addelement(self)
        
    def __repr__(self):
        return self.name + ' at ' + str((self.xc, self.yc))

    def initialize(self):
        self.aq = self.model.aq.find_aquifer_data(self.xc, self.yc)
        self.setbc()
        self.setflowcoef()
        # Since recharge is in layer 0, and RHS is -N
        self.an = self.aq.coef[0, :] * self.flowcoef  
        self.an.shape = (self.aq.naq, self.model.nint, self.model.npint)
        self.termin  = self.aq.lab2 * self.R * self.an
        self.termin2 = self.aq.lab2 ** 2 * self.an
        self.terminq = self.R * self.an
        self.termout = self.aq.lab2 * self.R * self.an
        self.i1R = iv(1, self.R / self.aq.lab2)
        self.k1R = kv(1, self.R / self.aq.lab2)
        self.termoutq= self.R * self.an
        self.dischargeinf = self.aq.coef[0, :] * self.flowcoef
        self.dischargeinflayers = np.sum(self.dischargeinf * 
                                         self.aq.eigvec[self.layers, :, :], 1)

    def setflowcoef(self):
        '''Separate function so that this can be overloaded for other types'''
        self.flowcoef = 1.0 / self.model.p  # Step function

    def potinf(self, x, y, aq=None):
        '''Can be called with only one x,y value'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        rv = np.zeros((self.nparam, aq.naq, self.model.nint,
                       self.model.npint), 'D')
        if aq == self.aq:
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            pot = np.zeros(self.model.npint, 'D')
            if r < self.R:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        #if r / abs(self.aq.lab2[i,j,0]) < self.rzero:
                        rv[0, i, j] = -self.termin[i, j] * \
                                      self.K1RI0r(r, i, j) + self.termin2[i, j]
            else:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        if (r - self.R) / \
                        abs(self.aq.lab2[i, j, 0]) < self.rzero:
                            rv[0, i, j, :] = self.termout[i, j, :] * \
                                             self.I1RK0r(r, i, j)                
        rv.shape = (self.nparam, aq.naq, self.model.npval)
        return rv
    
    def disvecinf(self,x,y,aq=None):
        '''Can be called with only one x,y value'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        qy = np.zeros((self.nparam, aq.naq, self.model.npval), 'D')
        if aq == self.aq:
            qr = np.zeros((self.nparam, aq.naq, self.model.nint, 
                           self.model.npint), 'D')
            r = np.sqrt((x - self.xc) ** 2 + (y - self.yc) ** 2)
            if r < self.R:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        #if r / abs(self.aq.lab2[i,j,0]) < self.rzero:
                        qr[0, i, j] = self.terminq[i, j] * self.K1RI1r(r, i, j)
            else:
                for i in range(self.aq.naq):
                    for j in range(self.model.nint):
                        if (r - self.R) / \
                        abs(self.aq.lab2[i, j, 0]) < self.rzero:
                            qr[0, i, j] = self.termoutq[i, j, :] * \
                                          self.I1RK1r(r, i, j)
            qr.shape = (self.nparam, aq.naq, self.model.npval)
            qx[:] = qr * (x - self.xc) / r
            qy[:] = qr * (y - self.yc) / r
        return qx, qy
    
    def plot(self):
        plt.plot(self.xc + self.R * np.cos(np.linspace(0, 2 * np.pi, 100)), \
                 self.yc + self.R * np.sin(np.linspace(0, 2 * np.pi, 100)), 'k')
        
    def K1RI0r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = np.sqrt(1 / (4 * r * R)) * np.exp(r - R) * \
            (1 + 3 / (8 * R) - 15 / (128 * R ** 2) + 315 / (3072 * R ** 3)) * \
            (1 + 1 / (8 * r) +  9 / (128 * r ** 2) + 225 / (3072 * r ** 3))
        else:
            rv = self.k1R[iaq, ipint] * iv(0, r)
        return rv
        
    def I1RK0r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = np.sqrt(1 / (4 * r * R)) * np.exp(R - r) * \
            (1 - 3 / (8 * R) - 15 / (128 * R ** 2) - 315 / (3072 * R ** 3)) * \
            (1 - 1 / (8 * r) +  9 / (128 * r ** 2) - 225 / (3072 * r ** 3))
        else:
            rv = self.i1R[iaq, ipint] * kv(0, r)
        return rv
    
    def K1RI1r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = np.sqrt(1 / (4 * r * R)) * np.exp(r - R) * \
            (1 + 3 / (8 * R) - 15 / (128 * R ** 2) + 315 / (3072 * R ** 3)) * \
            (1 - 3 / (8 * r) - 15 / (128 * r ** 2) - 315 / (3072 * r ** 3))
        else:
            rv = self.k1R[iaq, ipint] * iv(1, r)
        return rv

    def I1RK1r(self, rin, iaq, ipint):
        r = rin / self.aq.lab2[iaq, ipint]
        R = self.R / self.aq.lab2[iaq, ipint]
        if np.isinf(self.i1R[iaq, ipint]).any():
            rv = np.sqrt(1 / (4 * r * R)) * np.exp(R - r) * \
            (1 - 3 / (8 * R) - 15 / (128 * R ** 2) - 315 / (3072 * R ** 3)) * \
            (1 + 3 / (8 * r) - 15 / (128 * r ** 2) + 315 / (3072 * r ** 3))
        else:
            rv = self.i1R[iaq, ipint] * kv(1, r)
        return rv


import numpy as np
import inspect # Used for storing the input

class Element:
    def __init__(self, model, nparam=1, nunknowns=0, layers=0, \
                 tsandbc=[(0, 0)], type='z', name='', label=None):
        '''Types of elements
        'g': strength is given through time
        'v': boundary condition is variable through time
        'z': boundary condition is zero through time
        Definition of nlayers, Ncp, Npar, nunknowns:
        nlayers: Number of layers that the element is screened in, 
        as set in Element
        Ncp: Number of control points along the element
        nparam: Number of parameters, commonly nlayers * Ncp
        nunknowns: Number of unknown parameters, commonly zero or Npar
        '''
        self.model = model
        self.aq = None # Set in the initialization function
        self.nparam = nparam  # Number of parameters
        self.nunknowns = nunknowns
        self.layers = np.atleast_1d(layers)
        self.nlayers = len(self.layers)
        #
        tsandbc = np.atleast_2d(tsandbc).astype('d')
        tsandbc_error = "tsandQ or tsandh need to be 2D lists" + \
                        " or arrays, like [(0, 1), (2, 5), (8, 0)] "
        assert tsandbc.shape[1] == 2, tsandbc_error
        self.tstart, self.bcin = tsandbc[:,0] - self.model.tstart, tsandbc[:,1]
        if self.tstart[0] > 0:
            self.tstart = np.hstack((np.zeros(1), self.tstart))
            self.bcin = np.hstack((np.zeros(1), self.bcin))
        #
        # 'z' boundary condition thru time or 'v' boundary condition thru time
        self.type = type  
        self.name = name
        self.label = label
        if self.label is not None:
            assert self.label not in self.model.elementdict.keys(), \
                   "TTim error: label " + self.label + " already exists"
        self.rzero = 30
        
    def setbc(self):
        if len(self.tstart) > 1:
            self.bc = np.zeros_like(self.bcin)
            self.bc[0] = self.bcin[0]
            self.bc[1:] = self.bcin[1:] - self.bcin[:-1]
        else:
            self.bc = self.bcin.copy()
        self.ntstart = len(self.tstart)
        
    def initialize(self):
        '''Initialization of terms that cannot be initialized before other 
        elements or the aquifer is defined.
        As we don't want to require a certain order of entering elements, 
        these terms are initialized when Model.solve is called 
        The initialization class needs to be overloaded 
        by all derived classes'''
        pass
    
    def potinf(self, x, y, aq=None):
        '''Returns complex array of size (nparam, naq, npval)'''
        raise 'Must overload Element.potinf()'
    
    def potential(self, x, y, aq=None):
        '''Returns complex array of size (ngvbc, naq, npval)'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.parameters[:, :, np.newaxis, :] * \
                      self.potinf(x, y, aq), 1)
    
    def unitpotential(self, x, y, aq=None):
        '''Returns complex array of size (naq, npval)
        Can be more efficient for given elements'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.potinf(x, y, aq), 0)
    
    def unitpotentialone(self, x, y, jtime, aq=None):
        '''Returns complex array of size (naq, npval)
        Can be more efficient for given elements'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        return np.sum(self.potinfone(x, y, jtime, aq), 0)
    
    def disvecinf(self, x, y, aq=None):
        '''Returns 2 complex arrays of size (nparam, naq, npval)'''
        raise 'Must overload Element.disvecinf()'
    
    def disvec(self, x, y, aq=None):
        '''Returns 2 complex arrays of size (ngvbc, naq, npval)'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        return np.sum(self.parameters[:, :, np.newaxis, :] * qx, 1), \
               np.sum(self.parameters[:, :, np.newaxis, :] * qy, 1)
    
    def unitdisvec(self, x, y, aq=None):
        '''Returns 2 complex arrays of size (naq, npval)
        Can be more efficient for given elements'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        return np.sum(qx, 0), np.sum(qy, 0)
    
    # Functions used to build equations
    def potinflayers(self, x, y, layers=0, aq=None):
        '''layers can be scalar, list, or array. 
        returns array of size (len(layers),nparam,npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.potinf(x, y, aq)
        rv = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec, 2)
        # first axis needs to be the number of layers
        rv = rv.swapaxes(0, 1) 
        return rv[layers, :]
    
    def potentiallayers(self, x, y, layers=0, aq=None):
        '''Returns complex array of size (ngvbc, len(layers),npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.potential(x, y, aq)
        phi = np.sum(pot[:, np.newaxis, :, :] * aq.eigvec, 2)
        return phi[:, layers, :]
    
    def unitpotentiallayers(self, x, y, layers=0, aq=None):
        '''Returns complex array of size (len(layers), npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        pot = self.unitpotential(x, y, aq)
        phi = np.sum(pot[np.newaxis, :, :] * aq.eigvec, 1)
        return phi[layers, :]
    
    def disvecinflayers(self, x, y, layers=0, aq=None):
        '''layers can be scalar, list, or array. 
        returns 2 arrays of size (len(layers),nparam,npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvecinf(x, y, aq)
        rvx = np.sum(qx[:, np.newaxis, :, :] * aq.eigvec, 2)
        rvy = np.sum(qy[:, np.newaxis, :, :] * aq.eigvec, 2)
        # first axis needs to be the number of layers
        rvx = rvx.swapaxes(0, 1)
        rvy = rvy.swapaxes(0, 1)
        return rvx[layers, :], rvy[layers, :]
    
    def disveclayers(self, x, y, layers=0, aq=None):
        '''Returns 2 complex array of size (ngvbc, len(layers), npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.disvec(x, y, aq)
        rvx = np.sum(qx[:, np.newaxis, :, :] * aq.eigvec, 2)
        rvy = np.sum(qy[:, np.newaxis, :, :] * aq.eigvec, 2)
        return rvx[:, layers, :], rvy[:, layers, :]
    
    def unitdisveclayers(self, x, y, layers=0, aq=None):
        '''Returns complex array of size (len(layers), npval)
        only used in building equations'''
        if aq is None:
            aq = self.model.aq.find_aquifer_data(x, y)
        qx, qy = self.unitdisvec(x, y, aq)
        rvx = np.sum(qx[np.newaxis, :, :] * aq.eigvec, 1)
        rvy = np.sum(qy[np.newaxis, :, :] * aq.eigvec, 1)
        return rvx[layers, :], rvy[layers, :]
    
    # Other functions
    def discharge(self, t, derivative=0):
        """The discharge in each layer
        
        Parameters
        ----------
        t : scalar, list or array
            times at which discharge is computed.
            t must be ordered and tmin <= t <= tmax
        
        Returns
        -------
        array of discharges (nlayers,len(t))
            Discharge in each screen with zeros for layers that are not
            screened
            
        """
        # Could potentially be more efficient if s is pre-computed for 
        # all elements, but may not be worthwhile to store as it is quick now
        time = np.atleast_1d(t).astype('d')
        if (time[0] < self.model.tmin) or (time[-1] > self.model.tmax):
            print('Warning, some of the times are smaller than tmin or' + \
                  'larger than tmax; zeros are substituted')
        rv = np.zeros((self.nlayers, np.size(time)))
        if self.type == 'g':
            s = self.dischargeinflayers * self.model.p ** derivative
            for itime in range(self.ntstart):
                t =  time - self.tstart[itime]
                for i in range(self.nlayers):
                    rv[i] += self.bc[itime] * \
                             self.model.inverseLapTran(s[i], t)
        else:
            s = np.sum(self.parameters[:, :, np.newaxis, :] * 
                       self.dischargeinf, 1)
            s = np.sum(s[:, np.newaxis, :, :] * self.aq.eigvec, 2)
            s = s[:, self.layers, :] * self.model.p ** derivative
            for k in range(self.model.ngvbc):
                e = self.model.gvbclist[k]
                for itime in range(e.ntstart):
                    t = time - e.tstart[itime]
                    if t[-1] >= self.model.tmin:  # Otherwise all zero
                        for i in range(self.nlayers):
                            rv[i] += e.bc[itime] * \
                                     self.model.inverseLapTran(s[k, i], t)
        return rv
        
    def headinside(self, t):
        print("This function not implemented for this element")
        return
    
    def storeinput(self, frame):
        self.inputargs, _, _, self.inputvalues = inspect.getargvalues(frame)
        
    def write(self):
        rv = self.name + '(' + self.model.modelname + ',\n'
        for key in self.inputargs[2:]:  # The first two are ignored
            if isinstance(self.inputvalues[key],np.ndarray):
                rv += key + ' = ' + np.array2string(self.inputvalues[key],
                                                    separator=',') + ',\n'
            elif isinstance(self.inputvalues[key],str):                
                rv += key + " = '" + self.inputvalues[key] + "',\n"
            else:
                rv += key + ' = ' + str(self.inputvalues[key]) + ',\n'
        rv += ')\n'
        return rv
    
    def run_after_solve(self):
        '''function to run after a solution is completed.
        for most elements nothing needs to be done,
        but for strings of elements some arrays may need to be filled'''
        pass
    
    def plot(self):
        pass
        


from os import path

from setuptools import setup

version = {}
with open("ttim/version.py") as fp:
    exec(fp.read(), version)

l_d = ""
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    l_d = f.read()


setup(
    name="ttim",
    version=version["__version__"],
    description="Transient multi-layer AEM Model",
    long_description=l_d,
    long_description_content_type='text/markdown',
    author="Mark Bakker",
    author_email="markbak@gmail.com",
    url="https://github.com/mbakker7/ttim",
    license="MIT",
    packages=["ttim"],
    python_requires='>=3.7',
    install_requires=["numpy>=1.17", "scipy>=1.5", "numba>=0.5",
                      "matplotlib>=3.1", "lmfit>=1.0", "pandas>=1.1"],
    classifiers=['Topic :: Scientific/Engineering :: Hydrology'],
)
